"""
Module with classes for pure periodic systems (Crystals)
There are three main classes on this module: *KPoints* for describing a grid, list or path of points
in reciprocal space. Lattice for storing a manipulating cell vectors and computing the reciprocal lattice.
CrystalSymmetry for computing spacegroups, finding primitives and refining cells.

"""
from .lattice import Lattice
from .kpoints import KPoints
from .symmetry import CrystalSymmetry


# __all__ = filter(lambda s: not s.startswith('_'), dir())



from ..codes import CodeRun
from .input import SiestaInput
import os


class SiestaRun(CodeRun):

    def __init__(self, workdir, input_path, pseudo_path, pseudo_file=None, pseudo_list=None):
        CodeRun.__init__(self, executable='siesta', workdir=workdir, use_mpi=False)

        if pseudo_file is None and pseudo_list is None:
            raise ValueError("Declare either the list of pseudo names or a file that contain the list.")

        if pseudo_list is not None and not os.path.isdir(pseudo_path):
            raise ValueError('ERROR: Directory with Pseudo-potentials not found: %s' % pseudo_path)

        if pseudo_path is not None and not os.path.exists(input_path):
            raise ValueError('ERROR: File not found: %s' % input_path)

        self.workdir = workdir
        self.input_path = input_path
        self.stdin_filename = 'siesta.fdf'
        self.stdout_filename = 'siesta.out'
        self.stderr_filename = 'siesta.err'

        self.pseudo_path = os.path.abspath(pseudo_path)

        if os.path.exists(pseudo_file):
            rf = open(pseudo_file)
            self.pseudos = rf.read().split()

    def set_inputs(self):
        # Create a siesta input object
        si = SiestaInput(self.input_path)
        
        if not os.path.exists(self.workdir):
            os.mkdir(self.workdir)

        si.write(self.workdir+os.sep+'siesta.fdf')

        for i in self.pseudos:
            psf = self.workdir + os.sep + i + '.psf'
            if os.path.isfile(psf):
                os.remove(psf)
            os.symlink(self.pseudo_path + os.sep + i + '.psf', psf)

    def get_outputs(self):
        pass

    def is_finished(self):
        outputfile = self.workdir+os.sep+'siesta.out'
        if not os.path.isfile(outputfile):
            return False
        rf = open(outputfile)
        data = rf.read()
        if data[-4:] == 'Job completed\n':
            return True
        else:
            return False

    def is_successful(self):
        return self.is_finished()




import os
import pymongo
from pychemia import Structure
import gridfs
from pychemia.utils.computing import hashfile


class PyChemiaQueue:
    def __init__(self, name='Queue', host='localhost', port=27017, user=None, passwd=None, ssl=False, replicaset=None):
        """
        Creates a MongoDB client to 'host' with 'port' and connect it to the database 'name'.
        Authentication can be used with 'user' and 'password'

        :param name: (str) The name of the database
        :param host: (str) The host as name or IP
        :param port: (int) The number of port to connect with the server (Default is 27017)
        :param user: (str) The user with read or write permissions to the database
        :param passwd: (str/int) Password to authenticate the user into the server

        :return:
        """
        self.db_settings = {'name': name, 'host': host, 'port': port, 'user': user, 'passwd': passwd, 'ssl': ssl}
        self.name = name
        uri = 'mongodb://'
        if user is not None:
            uri += user
            if passwd is not None:
                uri += ':' + str(passwd)
            uri += '@'
        uri += host + ':' + str(port)
        print('URI:', uri)
        if user is not None:
            uri += '/' + name
        if replicaset is not None:
            self._client = pymongo.MongoClient(uri, ssl=ssl, replicaset=replicaset)
        else:
            self._client = pymongo.MongoClient(host=host, port=port, ssl=ssl,
                                               ssl_cert_reqs=pymongo.ssl_support.ssl.CERT_NONE)
        for i in ['version']:
            print('%20s : %s' % (i, self._client.server_info()[i]))
        self.db = self._client[name]
        if user is not None and self.db.authenticate(user, passwd):
            print('Authentication successful')

        self.set_minimal_schema()
        self.fs = gridfs.GridFS(self.db)

    def add_file(self, entry_id, location, filepath):

        assert (os.path.isfile(filepath))
        hashcode = hashfile(filepath)
        rf = open(filepath, 'rb')
        filename = os.path.basename(filepath)
        length = os.path.getsize(filepath)

        existing = self.db.fs.files.find_one({'hash': hashcode, 'length': length, 'filename': filename})

        if existing is None:

            file_id = self.fs.put(rf, filename=os.path.basename(filename), hash=hashcode)
            print('New file ', file_id)
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$addToSet': {location + '.files': {'file_id': file_id,
                                                                                                    'name': filename,
                                                                                                    'hash': hashcode}}})
        else:
            file_id = existing['_id']
            print('File already present ', file_id)
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$addToSet': {location + '.files': {'file_id': file_id,
                                                                                                    'name': filename,
                                                                                                    'hash': hashcode}}})

    def add_input_file(self, entry_id, filename):
        self.add_file(entry_id, 'input', filename)

    def set_minimal_schema(self):
        for entry in self.db.pychemia_entries.find({'meta': None}, {'_id': 1}):
            print('Missing field "meta" on', entry['_id'])
            self.db.pychemia_entries.update_one({'_id': entry['_id']}, {'$set': {'meta': {}}})
        for entry in self.db.pychemia_entries.find({'input': None}, {'_id': 1}):
            print('Missing field "input" on', entry['_id'])
            self.db.pychemia_entries.update_one({'_id': entry['_id']}, {'$set': {'input': {}}})
        for entry in self.db.pychemia_entries.find({'output': None}, {'_id': 1}):
            print('Missing field "output" on', entry['_id'])
            self.db.pychemia_entries.update_one({'_id': entry['_id']}, {'$set': {'output': {}}})
        for entry in self.db.pychemia_entries.find({'job': None}, {'_id': 1}):
            print('Missing field "job" on', entry['_id'])
            self.db.pychemia_entries.update_one({'_id': entry['_id']}, {'$set': {'job': {}}})

    def set_structure(self, entry_id, location, structure):
        assert (location in ['input', 'output'])
        self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {location + '.structure': structure.to_dict}})

    def set_input(self, entry_id, code, inputvar):

        for i in inputvar:
            if i.startswith('$'):
                value = inputvar.pop(i)
                inputvar[i[1:]] = value
        print('INPUTVAR:\n %s %s' % (inputvar, dict(inputvar)))
        self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'input.variables': dict(inputvar),
                                                                     'input.code': code.lower()}})

    def new_entry(self, structure=None, variables=None, code=None, files=None, priority=0, dbname=None, db_id=None):

        if variables is not None and code is None:
            raise ValueError("Input variables requiere code name")
        if variables is None and code is not None:
            raise ValueError("Input variables require code name")

        entry = {'input': {}, 'output': {}, 'job': {}, 'meta': {'submitted': False,
                                                                'priority': priority,
                                                                'finished': False,
                                                                'deployed': False,
                                                                'dbname': dbname,
                                                                'db_id': db_id}}
        result = self.db.pychemia_entries.insert_one(entry)
        entry_id = result.inserted_id

        # Commented for compatibility with mongo 2.4
        # self.db.pychemia_entries.update_one({'_id': entry_id}, {'$currentDate': {'meta.CreationDate': True}})

        if structure is not None:
            self.set_input_structure(entry_id, structure)
        if variables is not None and code is not None:
            self.set_input(entry_id, code=code, inputvar=variables)
        if files is not None:
            for ifile in files:
                self.add_input_file(entry_id, filename=ifile)
        return entry_id

    def set_job_settings(self, entry_id, nparal=None, queue=None, nhours=None, mail=None, task_name=None,
                         task_settings=None, task_kind=None):
        if nparal is not None:
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'job.nparal': nparal}})
        if queue is not None:
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'job.queue': queue}})
        if mail is not None:
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'job.mail': mail}})
        if nhours is not None:
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'job.nhours': nhours}})
        if task_name is not None:
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'job.task_name': task_name}})
        if task_kind is not None:
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'job.task_kind': task_name}})
        if task_settings is not None:
            self.db.pychemia_entries.update_one({'_id': entry_id}, {'$set': {'job.task_settings': task_settings}})

    def set_input_structure(self, entry_id, structure):
        return self.set_structure(entry_id, 'input', structure)

    def set_output_structure(self, entry_id, structure):
        return self.set_structure(entry_id, 'output', structure)

    def get_input_variables(self, entry_id):
        entry = self.db.pychemia_entries.find_one({'_id': entry_id}, {'input.variables': 1})
        if 'variables' in entry['input']:
            return entry['input']['variables']
        else:
            return None

    def get_structure(self, entry_id, location):
        """
        Return the structure in the entry with id 'entry_id'

        :rtype : Structure
        """
        assert (location in ['input', 'output'])
        entry = self.db.pychemia_entries.find_one({'_id': entry_id}, {location: 1})
        return Structure.from_dict(entry[location]['structure'])

    def get_input_structure(self, entry_id):
        return self.get_structure(entry_id, 'input')

    def get_output_structure(self, entry_id):
        return self.get_structure(entry_id, 'output')

    def write_input_files(self, entry_id, destination=None):

        if destination is None:
            dest = '.'
        elif os.path.isfile(destination):
            dest = os.path.dirname(os.path.abspath(destination))
        elif not os.path.exists(destination):
            os.mkdir(destination)
            dest = destination
        elif os.path.isdir(destination):
            dest = destination
        else:
            raise ValueError('Destination not valid')

        entry = self.db.pychemia_entries.find_one({'_id': entry_id}, {'input.files': 1})
        for ifile in entry['input']['files']:
            rf = self.fs.get(ifile['file_id'])
            wf = open(dest + os.sep + ifile['name'], 'wb')
            wf.write(rf.read())
            wf.close()



import os
import json
import numpy as np
from pychemia.crystal import KPoints
from ...tasks import Task
from ..abinit import AbinitJob

__author__ = 'Guillermo Avendano-Franco'


class StaticCalculation(Task):
    def __init__(self, structure, workdir='.', executable='abinit', ecut=50, kpoints=None, kp_density=1E4):

        self.ecut = ecut
        if kpoints is None:
            kp = KPoints.optimized_grid(structure.lattice, kp_density=kp_density, force_odd=True)
            self.kpoints = kp
        else:
            self.kpoints = kpoints
        self.task_params = {'ecut': self.ecut, 'kpoints': self.kpoints.to_dict}
        Task.__init__(self, structure=structure, task_params=self.task_params, workdir=workdir, executable=executable)
        self.abinitjob = AbinitJob()
        self.abinitjob.initialize(structure=structure)
        self.is_prepared=False

    def prepare(self):
        self.abinitjob.set_kpoints(kpoints=self.kpoints)
        self.abinitjob.job_static()
        self.abinitjob.set_ecut(self.ecut)
        self.abinitjob.set_psps()
        self.abinitjob.write_all()
        self.is_prepared = True

    def run(self, num_threads=None, mpi_num_procs=None):
        if not self.is_prepared:
            self.prepare()
        self.abinitjob.run(num_threads=num_threads, mpi_num_procs=mpi_num_procs)

    def plot(self, figname='static_calculation.pdf'):
        if not self.finished:
            print('The task is not finished')
            return
        import matplotlib.pyplot as plt
        plt.switch_backend('agg')
        plt.figure(figsize=(8, 6))
        plt.subplots_adjust(left=0.09, bottom=0.08, right=0.95, top=0.95, wspace=None, hspace=None)
        data = np.array(self.output['energies'])
        plt.plot(data[:, 1], data[:, 2], 'b.-')
        plt.xlabel('SCF cycle')
        plt.ylabel('Energy [eV]')

        a = plt.axes([.6, .6, .3, .3], axisbg='0.9')
        a.semilogy(data[:, 1], data[:, 2] - np.min(data[:, 2]))
        a.set_title('min energy %7.3f eV' % np.min(data[:, 2]))

        if figname is not None:
            plt.savefig(figname)
        return plt.gcf()

    def load(self, filename=None):
        if filename is None:
            filename = self.workdir + os.sep + 'task.json'
        rf = open(filename)
        data = json.load(rf)
        rf.close()
        self.task_params = data['task_params']
        self.output = data['output']
        self.ecut = self.task_params['ecut']
        self.kpoints = KPoints.from_dict(self.task_params['kpoints'])

    def report(self, file_format='html'):
        from lxml.builder import ElementMaker, E
        self.plot(figname=self.report_dir + os.sep + 'static.jpg')

        element_maker = ElementMaker(namespace=None, nsmap={None: "http://www.w3.org/1999/xhtml"})
        html = element_maker.html(E.head(E.title("ABINIT Static Calculation")),
                                  E.body(E.h1("ABINIT Static Calculation"),
                                         E.h2('Structure'),
                                         E.pre(str(self.structure)),
                                         E.h2('Self Consistent Field Convergence'),
                                         E.p(E.img(src='static.jpg', width="800", height="600",
                                                   alt="Static Calculation"))
                                         ))

        return self.report_end(html, file_format)



"""
Routines to import and export atomic structures.
Currently supported are *ASCII*, *CIF* and *XYZ*
"""

from . import cif
from . import ascii
from . import xyz

# __all__ = filter(lambda s: not s.startswith('_'), dir())



import random
from .searcher import Searcher
from pychemia import pcm_log


class HarmonySearch(Searcher):
    def __init__(self, population, params=None, generation_size=32, stabilization_limit=10):
        """
        Harmony Search Method:
        This searcher is the simplest one, does not require a metric space except for the evaluation
        of duplicates. It is more elaborated than a simple random search.
        Harmony Search only requires changers for existing members and generation of new random members.

        :param population: A subclass of the ABC Population
        :param params: (dict) Parameters to setup the Searcher
                    hmcr: Harmony Memory Considering Rate, how often a value should be discarded in favor of new random
                          members, a value of 1 means never discard any member except those in tail (see below).
                          A value of 0 means always discard except those in tail (see below).
                    par:  Probability of being changed for those not discarded.
                          A value of 1 means always changed, a value of 0 means never changed.
                    top:  Number of members that are automatically promoted to the next generation (The best)
                    tail: Number of members that are automatically discarded (The worst)
        :param stabilization_limit: Number of generations survived by the best member
        """
        # Mandatory objects
        Searcher.__init__(self, population, generation_size, stabilization_limit)
        # Parameters
        self.hmcr = 0.9  # harmony_memory_considering_rate
        self.par = 0.9
        self.top = 2
        self.tail = 2
        self.set_params(params)
        self.delta_change = 1

    def set_params(self, params):
        """
        Set the parameters for the Harmony Search algorithm
        There are 4 parameters

        :param params:
        """
        if params is None:
            params = {}
        if 'hmcr' in params:
            self.hmcr = params['hmcr']
        if 'par' in params:
            self.par = params['par']
        if 'top' in params:
            self.top = params['top']
        if 'tail' in params:
            self.tail = params['tail']
        assert (1.0 >= self.hmcr >= 0.0)
        assert (1.0 >= self.par >= 0.0)

    def get_params(self):
        return {'hmcr': self.hmcr, 'par': self.par}

    def run_one(self):
        """
        Run one cycle for the Harmony Search Method
        """
        # Get a static selection of the values in the generation that are evaluated
        selection = self.population.ids_sorted(self.actives_in_generation)
        pcm_log.debug(' Size of selection : %d' % len(selection))

        # Automatic promotion for the top ranking members
        for entry_id in selection[:min(self.top, len(selection))]:
            pcm_log.debug('[HS](%s) Top entry: promoted' % str(entry_id))
            self.pass_to_new_generation(entry_id, reason='Top %d' % self.top)
        # assert(len(self.get_generation(self.current_generation+1)) >= 2)

        if len(selection) > self.top + self.tail:
            # Intermediate members, their fate depends on hmcr and par
            for entry_id in selection[self.top:-self.tail]:
                rnd = random.random()
                pcm_log.debug('[HS](%s) Middle entry: rnd=%4.3f hmcr= %4.3f' % (entry_id, rnd, self.hmcr))
                if rnd <= self.hmcr:
                    rnd = random.random()
                    if rnd < self.par:
                        pcm_log.debug('[HS](%s) Promoted (modified): rnd= %4.3f < par= %4.3f' % (entry_id,
                                                                                                 rnd,
                                                                                                 self.par))
                        self.replace_by_changed(entry_id, reason='rnd= %4.3f < par= %4.3f' % (rnd, self.par))
                    else:
                        pcm_log.debug('[HS](%s) Promoted (unmodified): rnd= %4.3f >= par= %4.3f' %
                                      (entry_id, rnd, self.par))
                        self.pass_to_new_generation(entry_id, reason='rnd= %4.3f >= par= %4.3f' % (rnd, self.par))
                else:
                    pcm_log.debug('[HS](%s) Discarded: rnd= %4.3f > hmcr= %4.3f ' % (entry_id, rnd, self.hmcr))
                    self.replace_by_random(entry_id, reason='rnd= %4.3f > hmcr= %4.3f' % (rnd, self.hmcr))

        tail = min(len(selection)-self.top, self.tail)
        if len(selection) > self.top:
            for entry_id in selection[-tail:]:
                pcm_log.debug('[HS](%s) Tail entry: discarded' % entry_id)
                self.replace_by_random(entry_id, reason='Tail %d' % self.tail)
                # assert(len(self.get_generation(self.current_generation+1)) == self.generation_size)




import json
from abc import ABCMeta, abstractmethod
import numpy as np
from pychemia.utils.computing import deep_unicode
from numbers import Integral, Real


class PyChemiaJsonable(object):
    """
    Abstract base class specifying how to convert objects from/to dictionaries.
    PyChemiaJsonable objects must implement a to_dict property and a from_dict static method.
    """
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def to_dict(self):
        """
        A JSON representation of an object.
        """
        pass

    @classmethod
    def from_dict(cls, json_dict):
        """
        This implements a default from_dict method which supports all
        classes that simply try to recreate an object using the keys
        as arguments for the creation of the new object.

        :param json_dict: Recreate an object from its serialize form
        :return:
        """
        argstring = ''
        for key in json_dict:
            argstring += key + '=' + str(json_dict[key]) + ', '
        argstring = argstring[:-2]
        print(str(cls) + '(' + argstring + ')')
        return eval(str(cls) + '(' + argstring + ')')

    @property
    def to_json(self):
        """
        Returns a json string representation of the object.
        """
        return json.dumps(self)

    def save_to_file(self, filename):
        """
        Writes the json representation to a file.

        :param filename: (str) Filename for the json that will be created
        """
        with open(filename, "w") as f:
            json.dump(self, f)


def generic_serializer(value):
    """
    A generic serializer for very common values

    :param value:
    :return:
    """
    value = deep_unicode(value)

    if value is None:
        return None
    elif isinstance(value, dict):
        new_value = {}
        for i in value:
            new_value[i] = generic_serializer(value[i])
        return new_value
    elif hasattr(value, '__iter__'):
        return [generic_serializer(element) for element in value]
    elif isinstance(value, str):
        return value
    elif isinstance(value, Integral):
        return int(value)
    elif isinstance(value, Real):
        return float(value)
    elif isinstance(value, np.integer):
        return int(value)
    elif isinstance(value, np.float):
        return float(value)
    else:
        raise ValueError("Could not serialize this: %s of type: %s" % (value, type(value)))



"""
Routines related to Data mining
"""

from pychemia import HAS_ASE, HAS_SPGLIB

if HAS_ASE and HAS_SPGLIB:
    from ._ase import ase2pychemia, pychemia2ase

# __all__ = filter(lambda s: not s.startswith('_'), dir())



import random
import itertools
import numpy as np
from pychemia.utils.mathematics import apply_rotation


class StructureChanger:
    def __init__(self, structure):

        self.old_structure = structure
        self.new_structure = structure.copy()

        self.operations = []

    def permutator(self, pair):

        self.new_structure.symbols[pair[0]] = self.old_structure.symbols[pair[1]]
        self.new_structure.symbols[pair[1]] = self.old_structure.symbols[pair[0]]

        self.operations.append({'permutator': (pair[0], pair[1])})

    def random_permutator(self, forbidden_list=None):

        assert (self.old_structure.nspecies > 1)
        pairs = list(itertools.combinations(self.old_structure.species, 2))

        if forbidden_list is None:
            flist = []
        else:
            flist = forbidden_list

        for pair in pairs:
            for fpair in flist:
                if pair == tuple(fpair) or pair == tuple(reversed(fpair)):
                    pairs.remove(pair)

        pair = random.choice(pairs)

        indices0 = [i for i, x in enumerate(self.old_structure.symbols) if x == pair[0]]
        index0 = random.choice(indices0)

        indices1 = [i for i, x in enumerate(self.old_structure.symbols) if x == pair[1]]
        index1 = random.choice(indices1)

        self.permutator((index0, index1))

    def deform_cell(self, stress_eps):

        stress = np.eye(3) + np.diag(stress_eps[:3]) + np.array([[0.0, stress_eps[3], stress_eps[4]],
                                                                 [stress_eps[3], 0.0, stress_eps[5]],
                                                                 [stress_eps[4], stress_eps[5], 0.0]])

        self.new_structure.set_cell(np.dot(stress, self.old_structure.cell))

        self.operations.append({'deform_cell': stress_eps})

    def random_deform_cell(self, diag=True, nondiag=True, maxdelta=0.01):

        # 6 random numbers between -maxdelta and maxdelta
        stress_eps = np.random.random(6) * 2 * maxdelta - maxdelta

        if diag:
            stress_eps[:3] = 0
        if nondiag:
            stress_eps[-3:] = 0

        self.deform_cell(stress_eps)

    def move_one_atom(self, index, vector):

        self.new_structure.positions[index] += vector
        self.new_structure.positions2reduced()

        self.operations.append({'move_one_atom': (index, vector)})

    def random_move_one_atom(self, mu=0.1, sigma=0.01):

        index = random.randint(0, len(self.old_structure) - 1)
        radius = np.abs(np.random.normal(mu, sigma))
        theta_x = 2 * np.pi * np.random.random_sample()
        theta_y = 2 * np.pi * np.random.random_sample()
        theta_z = 2 * np.pi * np.random.random_sample()
        vector = apply_rotation([1, 0, 0], theta_x, theta_y, theta_z)

        self.move_one_atom(index, vector * radius)

    def random_move_many_atoms(self, epsilon=0.01, forbidden_indices=None, forbidden_species=None):

        if forbidden_indices is None:
            findices = []
        else:
            findices = forbidden_indices

        if forbidden_species is None:
            fspec = []
        else:
            fspec = forbidden_species

        for iatom in range(len(self.old_structure)):
            if iatom not in findices and self.new_structure.symbols[iatom] not in fspec:
                self.random_move_one_atom(epsilon)

    def random_change(self, epsilon):
        rnd = np.random.random()
        if rnd < 0.25:
            self.random_deform_cell(maxdelta=epsilon)
        elif rnd < 0.5:
            self.random_move_many_atoms(epsilon=epsilon)
        elif rnd < 0.75 and self.old_structure.nspecies > 1:
            self.random_permutator()
        else:
            self.random_move_one_atom(mu=epsilon, sigma=0.01)



"""
Created on Mon Dec 02 2019 
@author: Pedram Tavadze
"""
#import pychemia                     
import xml.etree.ElementTree as ET    
from numpy import array 


def text_to_bool(text):
    """boolians in vaspxml are stores as T or F in str format, this function coverts them to python boolians """
    text = text.strip(' ')
    if text == 'T' or text == '.True.' or text == '.TRUE.':
        return True
    else:
        return False


def conv(ele, _type):
    """This function converts the xml text to the type specified in the attrib of xml tree """
    
    if _type == 'string':
        return ele.strip()
    elif _type == 'int':
        return int(ele) 
    elif _type == 'logical': 
        return text_to_bool(ele)
    elif _type == 'float':
        return float(ele) 


def get_varray(xml_tree):
    """Returns an array for each varray tag in vaspxml """
    ret = [] 
    for ielement in xml_tree: 
        ret.append([float(x) for x in ielement.text.split()]) 
    return ret


def get_params(xml_tree, dest):
    """dest should be a dictionary
    This function is recurcive #check spelling"""
    for ielement in xml_tree:
        if ielement.tag == 'separator':
            dest[ielement.attrib['name'].strip()] = {} 
            dest[ielement.attrib['name'].strip()] = get_params(ielement, dest[ielement.attrib['name']]) 
        else : 
            if 'type' in ielement.attrib:
                _type = ielement.attrib['type']
            else : 
                _type = 'float' 
            if ielement.text is None:
                dest[ielement.attrib['name'].strip()] = None 
                
            elif len(ielement.text.split()) > 1:
                dest[ielement.attrib['name'].strip()] = [conv(x, _type) for x in ielement.text.split()]
            else : 
                dest[ielement.attrib['name'].strip()] = conv(ielement.text, _type)

    return dest 


def get_structure(xml_tree):
    """Returns a dictionary of the structure """
    ret = {}
    for ielement in xml_tree:
        if ielement.tag == 'crystal': 
            for isub in ielement: 
                if isub.attrib['name'] == 'basis': 
                    ret['cell'] = get_varray(isub)
                elif isub.attrib['name'] == 'volume': 
                    ret['volume'] = float(isub.text) 
                elif isub.attrib['name'] == 'rec_basis': 
                    ret['rec_cell'] = get_varray(isub)
        elif ielement.tag == 'varray':
            if ielement.attrib['name'] == 'positions': 
                ret['reduced'] = get_varray(ielement)
    return ret


def get_scstep(xml_tree):
    """This function extracts the self-consistent step information """
    scstep = {'time': {}, 'energy': {}}
    for isub in xml_tree:
        if isub.tag=='time': 
            scstep['time'][isub.attrib['name']] = [float(x) for x in isub.text.split()] 
        elif isub.tag == 'energy': 
            for ienergy in isub:
                scstep['energy'][ienergy.attrib['name']] = float(ienergy.text) 
    return scstep


def get_set(xml_tree, ret):
    """ This function will extract any element taged set recurcively"""
    if xml_tree[0].tag == 'r': 
        ret[xml_tree.attrib['comment']] = get_varray(xml_tree)         
        return ret 
    else:
        ret[xml_tree.attrib['comment']] = {}
        for ielement in xml_tree:
            
            if ielement.tag == 'set' :   
                ret[xml_tree.attrib['comment']][ielement.attrib['comment']] = {}
                ret[xml_tree.attrib['comment']][ielement.attrib['comment']] = get_set(ielement, ret[xml_tree.attrib['comment']][ielement.attrib['comment']])
        return ret 


def get_general(xml_tree, ret):
    """ This function will parse any element in calculatio other than the structures, scsteps"""
    if 'dimension' in [x.tag for x in xml_tree]: 
        ret['info'] = [] 
        ret['data'] = {} 
        for ielement in xml_tree: 
            if ielement.tag == 'field': 
                ret['info'].append(ielement.text.strip(' ')) 
            elif ielement.tag == 'set': 
                for iset in ielement: 
                    ret['data'] = get_set(iset, ret['data'])
        return ret 
    else : 
        for ielement in xml_tree: 
            if ielement.tag == 'i': 
                if 'name' in ielement.attrib:
                    if ielement.attrib['name'] == 'efermi':
                        ret['efermi'] = float(ielement.text)
                continue 
            ret[ielement.tag] = {} 
            ret[ielement.tag] = get_general(ielement, ret[ielement.tag]) 
        return ret   


def parse_vasprun(vasprun):
    tree = ET.parse(vasprun)
    root = tree.getroot()

    calculation = []
    structures = []
    forces = []
    stresses = []
    orbital_magnetization = {}
    run_info = {}
    incar = {}
    general = {}
    kpoints_info = {}
    vasp_params = {}
    kpoints_list = [] 
    k_weights = []
    atom_info = {} 
    for ichild in root:

        if ichild.tag == 'generator':
            for ielement in ichild:
                run_info[ielement.attrib['name']] = ielement.text

        elif ichild.tag == 'incar':
            incar = get_params(ichild, incar)
    
        ## Skipping 1st structure which is primitive cell
        elif ichild.tag == 'kpoints':
    
            for ielement in ichild:
                if ielement.items()[0][0] == 'param': 
                    kpoints_info['mode'] = ielement.items()[0][1] 
                    if kpoints_info['mode'] == 'listgenerated':
                        kpoints_info['kpoint_vertices'] = []
                        for isub in ielement:
                            
                            if isub.attrib == 'divisions':
                                kpoints_info['ndivision'] = int(isub.text)
                            else:
                                if len(isub.text.split()) !=3:
                                    continue
                                kpoints_info['kpoint_vertices'].append([float(x) for x in isub.text.split()])
                    else : 
                        for isub in ielement:
                            if isub.attrib['name'] == 'divisions': 
                                kpoints_info['kgrid'] = [int(x) for x in isub.text.split()] 
                            elif isub.attrib['name'] == 'usershift': 
                                kpoints_info['user_shift'] = [float(x) for x in isub.text.split()] 
                            elif isub.attrib['name'] == 'genvec1': 
                                kpoints_info['genvec1'] = [float(x) for x in isub.text.split()] 
                            elif isub.attrib['name'] == 'genvec2': 
                                kpoints_info['genvec2'] = [float(x) for x in isub.text.split()] 
                            elif isub.attrib['name'] == 'genvec3': 
                                kpoints_info['genvec3'] = [float(x) for x in isub.text.split()] 
                            elif isub.attrib['name'] == 'shift': 
                                kpoints_info['shift'] = [float(x) for x in isub.text.split()] 
    
                elif ielement.items()[0][1] == 'kpointlist': 
                    for ik in ielement:
                        kpoints_list.append([float(x) for x in ik.text.split()]) 
                    kpoints_list = array(kpoints_list)
                elif ielement.items()[0][1] == 'weights': 
                    for ik in ielement:
                        k_weights.append(float(ik.text))
                    k_weights = array(k_weights)
                    
        ## Vasp Parameters 
        elif ichild.tag == 'parameters':
            vasp_params = get_params(ichild,vasp_params) 
    
        ## Atom info
        elif ichild.tag == 'atominfo':

            for ielement in ichild:
                if ielement.tag == 'atoms':
                    atom_info['natom'] = int(ielement.text) 
                elif ielement.tag == 'types':
                    atom_info['nspecies'] = int(ielement.text) 
                elif ielement.tag == 'array':
                    if ielement.attrib['name'] == 'atoms': 
                        for isub in ielement:
                            if isub.tag == 'set': 
                                atom_info['symbols'] = [] 
                                for isym  in isub : 
                                    atom_info['symbols'].append(isym[0].text) 
                    elif ielement.attrib['name'] == 'atomtypes': 
                        atom_info['atom_types'] = {} 
                        for isub in ielement:
                            if isub.tag == 'set': 
                                for iatom in isub:
                                    atom_info['atom_types'][iatom[1].text] = {} 
                                    atom_info['atom_types'][iatom[1].text]['natom_per_specie'] = int(iatom[0].text) 
                                    atom_info['atom_types'][iatom[1].text]['mass'] = float(iatom[2].text) 
                                    atom_info['atom_types'][iatom[1].text]['valance'] = float(iatom[3].text) 
                                    atom_info['atom_types'][iatom[1].text]['pseudopotential'] = iatom[4].text.strip()

        elif ichild.tag == 'structure':
            if ichild.attrib['name'] == 'initialpos': 
                initial_pos = get_structure(ichild)
            elif ichild.attrib['name'] == 'finalpos': 
                final_pos = get_structure(ichild)

        elif ichild.tag == 'calculation':
            for ielement in ichild:
                if  ielement.tag == 'scstep': 
                    calculation.append(get_scstep(ielement)) 
                elif ielement.tag == 'structure':
                    structures.append(get_structure(ielement))
                elif ielement.tag == 'varray': 
                    if ielement.attrib['name'] == 'forces': 
                        forces.append(get_varray(ielement)) 
                    elif ielement.attrib['name'] == 'stress': 
                        stresses.append(get_varray(ielement)) 
                        
                # elif ielement.tag == 'eigenvalues': 
                #     for isub in ielement[0] :  
                #         if isub.tag == 'set': 
                #             for iset in isub : 
                #                 eigen_values[iset.attrib['comment']] = {} 
                #                 for ikpt in iset :  
                #                     eigen_values[iset.attrib['comment']][ikpt.attrib['comment']] = get_varray(ikpt) 
                
                elif ielement.tag == 'separator': 
                    if ielement.attrib['name'] == "orbital magnetization": 
                        for isub in ielement:
                            orbital_magnetization[isub.attrib['name']] = [float(x) for x in isub.text.split()]

                # elif ielement.tag == 'dos': 
                #     for isub in ielement :  
                #         if 'name' in isub.attrib: 
                #             if isub.attrib['name'] == 'efermi' :  
                #                 dos['efermi'] = float(isub.text) 
                #             else :  
                #                 dos[isub.tag] = {} 
                #                 dos[isub.tag]['info'] = [] 
                #               for iset in isub[0]  : 
                #                   if iset.tag == 'set' : 
                #                       for isub_set in iset: 
                #                           dos[isub.tag] = get_set(isub_set,dos[isub.tag]) 
                #                   elif iset.tag == 'field' : 
                #                       dos[isub.tag]['info'].append(iset.text.strip(' '))
                else:
                    general[ielement.tag] = {}
                    general[ielement.tag] = get_general(ielement, general[ielement.tag])
        # NEED TO ADD ORBITAL MAGNETIZATION 

    return {'calculation': calculation, 'structures': structures, 'forces': forces, 'run_info': run_info,
            'incar': incar, 'general': general, 'kpoints_info': kpoints_info, 'vasp_params': vasp_params,
            'kpoints': {'kpoints_list': kpoints_list, 'k_weights': k_weights}, 'atom_info': atom_info}






#!D:\bld\pychemia_1604228883768\_h_env\python.exe

import logging
import os
import sys
import json
import getopt
import pychemia
from pychemia.code.vasp.task import ConvergenceKPointGrid, IonRelaxation, ConvergenceCutOffEnergy
from pychemia.utils.computing import get_float, get_int

logging.basicConfig(level=logging.DEBUG)

try:
    from pychemia.symm import symmetrize
except ImportError:
    def symmetrize(x):
        return x


def cleaner():
    files = ['DOSCAR', 'EIGENVAL', 'IBZKPT', 'OSZICAR', 'PCDAT', 'PROCAR', 'WAVECAR', 'XDATCAR']
    [os.remove(ifile) for ifile in files if os.path.exists(ifile)]


def usage(name):
    print("""
NAME
    %s

DESCRIPTION
    Vasp relaxator with convergence in K-Point Grid and Cut-Off Energy (ENCUT)

OPTIONS

    --help, -h
        Return information on the options and use of this script

    --structure, -s <string> (Default: 'POSCAR')
        Structure for which the strength will be computed

    --output, -o <string> (Default: 'pcm_results.json')
        File in JSON format where the results are stored

    --nparal, -n <int> (Default: 2)
        Number of MPI parallel processes for VASP

    --binary, -b <string> (Default: 'vasp')
        Path to the VASP executable

    --energy_tol, -e <float> (Default: 1E-3)
        Energy tolerance for three succesive values of the Kpoint grid or CutOff energy

    --target_forces, -t <float> (Default: 1E-3)
        Target forces for internal relaxation
""" % os.path.basename(name))


def main(argv):
    try:
        opts, args = getopt.getopt(argv[1:], "hs:o:n:b:e:t:", ["help", "structure=", "output=", "nparal=",
                                                               "binary=", "energy_tol=", "target_forces="])
    except getopt.GetoptError:
        usage(argv[0])
        sys.exit(2)

    if len(opts) == 0:
        usage(argv[0])
        sys.exit(2)

    # Default Values
    workdir = '.'
    structure_file = 'POSCAR'
    output_file = 'pcm_results.json'
    nparal = 2
    energy_tol = 1E-3
    target_forces = 1E-3
    binary = 'vasp'

    for opt, arg in opts:

        if opt in ("-h", "--help"):
            usage(argv[0])
            sys.exit()
        elif opt in ("-s", "--structure"):
            structure_file = arg
        elif opt in ("-o", "--output"):
            output_file = arg
        elif opt in ("-n", "--nparal"):
            nparal = get_int(arg)
        elif opt in ("-b", "--binary"):
            binary = arg
        elif opt in ("-e", "--energy_tol"):
            energy_tol = get_float(arg)
        elif opt in ("-t", "--target_forces"):
            target_forces = get_float(arg)
        elif opt in ("-s", "--structure"):
            structure_file = arg

    structure = pychemia.structure_from_file(structure_file)
    if structure is None:
        print(" ERROR: Invalid structure, no structure could be obtained from '%s'" % structure_file)
        sys.exit(2)

    if structure_file == 'POSCAR':
        os.rename('POSCAR', 'POSCAR_original')

    print("\n PyChemia VASP Relaxator")
    print(" =======================\n")
    print(" VASP binary         : ", binary)
    print(" Energy tolerance    : ", energy_tol)
    print(" Target forces       : ", target_forces)
    print(" MPI number of procs : ", nparal)
    print(" Structure           :\n")
    print(structure)

    wf = open(output_file, 'w')
    data = {'input': {'binary': binary,
                      'energy_tol': energy_tol,
                      'target_forces': target_forces,
                      'nparal': nparal,
                      'structure': structure.to_dict}}
    json.dump(data, wf)
    wf.close()

    # First Round (Relaxing the original structure)
    print('\nFirst Round')
    print('===========')

    cleaner()
    print('\nConvergence of Cut-off Energy')
    print('-----------------------------\n')
    ce = ConvergenceCutOffEnergy(structure, energy_tolerance=energy_tol)
    if os.path.isfile('convergence_encut.json'):
        print('A previous convergence study was found, loading...')
        ce.load()
    if not ce.is_converge:
        ce.run(nparal)
        ce.save()
        ce.plot()
    encut = ce.best_encut

    data['output'] = {'1R_ENCUT': encut}
    wf = open(output_file, 'w')
    json.dump(data, wf)
    wf.close()

    cleaner()
    print('\nConvergence of K-Point Grid')
    print('---------------------------\n')
    ck = ConvergenceKPointGrid(structure, encut=encut, energy_tolerance=energy_tol)
    if os.path.isfile('convergence_kpoints.json'):
        print('A previous convergence study was found, loading...')
        ce.load()
    if not ce.is_converge:
        ce.run(nparal)
        ce.save()
        ce.plot()
    kp = ck.best_kpoints

    data['output'] = {'1R_KPOINTS': list(kp.grid)}
    print(data)
    wf = open(output_file, 'w')
    json.dump(data, wf)
    wf.close()

    os.rename('convergence_encut.json', 'convergence_encut_phase1.json')
    os.rename('convergence_encut.pdf', 'convergence_encut_phase1.pdf')
    os.rename('convergence_kpoints.json', 'convergence_kpoints_phase1.json')
    os.rename('convergence_kpoints.pdf', 'convergence_kpoints_phase1.pdf')

    cleaner()
    print('\nIonic Relaxation')
    print('----------------\n')
    vr = IonRelaxation(structure=structure, encut=encut, kp_grid=kp.grid, workdir=workdir,
                       target_forces=10 * target_forces)
    vr.run(nparal)

    structure = vr.get_final_geometry()
    structure.save_json(workdir + os.sep + 'structure_phase1.json')

    data['output'] = {'1R_structure': structure.to_dict}
    wf = open(output_file, 'w')
    json.dump(data, wf)
    wf.close()

    # Second Round (Symetrize structure and redo convergences)
    st = symmetrize(structure)

    print('\nSecond Round')
    print('============')

    cleaner()
    print('\nConvergence of K-Point Grid')
    print('---------------------------\n')
    ck = ConvergenceKPointGrid(st, encut=encut, energy_tolerance=energy_tol, recover=True)
    ck.run(nparal)
    ck.save()
    ck.plot()
    kp = ck.best_kpoints

    data['output'] = {'2R_KPOINTS': list(kp.grid)}
    wf = open(output_file, 'w')
    json.dump(data, wf)
    wf.close()

    cleaner()
    print('\nConvergence of Cut-off Energy')
    print('-----------------------------\n')
    ce = ConvergenceCutOffEnergy(st, energy_tolerance=energy_tol, kpoints=kp)
    ce.run(nparal)
    ce.save()
    ce.plot()
    encut = ce.best_encut

    data['output'] = {'2R_ENCUT': encut}
    wf = open(output_file, 'w')
    json.dump(data, wf)
    wf.close()

    os.rename('convergence_encut.json', 'convergence_encut_phase2.json')
    os.rename('convergence_encut.pdf', 'convergence_encut_phase2.pdf')
    os.rename('convergence_kpoints.json', 'convergence_kpoints_phase2.json')
    os.rename('convergence_kpoints.pdf', 'convergence_kpoints_phase2.pdf')

    cleaner()
    print('\nIonic Relaxation')
    print('----------------\n')
    vr = IonRelaxation(structure=st, workdir='.', encut=encut, kp_grid=kp.grid, target_forces=target_forces)
    vr.run(nparal)

    structure = vr.get_final_geometry()
    structure.save_json(workdir + os.sep + 'structure_phase2.json')

    data['output'] = {'2R_structure': structure.to_dict}

    wf = open(output_file, 'w')
    json.dump(data, wf)
    wf.close()

    cleaner()


if __name__ == "__main__":
    main(sys.argv)



import os
import numpy as np
from scipy.io.netcdf import netcdf_file


def file2dict(filename):

    if not os.path.isfile(filename):
        raise ValueError("ERROR: Could not read %s" % filename)

    nc = netcdf_file(filename, 'r', mmap=False)
    ret = {}
    for ikey in nc.variables.keys():
        data = nc.variables[ikey].data
        if type(data[0]) == np.float64:
            if len(data) == 1:
                data = float(data[0])
            else:
                data = [float(x) for x in data]
        elif type(data[0]) == np.int32:
            if len(data) == 1:
                data = int(data[0])
            else:
                data = [int(x) for x in data]
        else:
            data = list(data)

        ret[ikey] = data
        del data

    nc.close()
    return ret


def netcdf2dict(filename):
    """
    Read a NetCDF file and create a python dictionary with
    numbers or lists for each variable

    Args:
        filename:
            NetCDF filename
    """
    if not os.path.isfile(filename):
        print('ERROR: No such file: ', filename)
        return None
    ret = {}
    netcdfile = netcdf_file(filename, 'r', mmap=False)
    for ii in netcdfile.variables.keys():
        ret[ii] = netcdfile.variables[ii][:]
    netcdfile.close()

    for i in ret:
        if ret[i].dtype == np.dtype('>f8'):
            ret[i] = [round(x, 11) for x in ret[i].flatten()]
        elif ret[i].dtype == np.dtype('>i4'):
            ret[i] = [int(x) for x in ret[i].flatten()]

    for i in ret:
        if len(ret[i]) == 1:
            ret[i] = ret[i][0]

    return ret



import numpy as np
from pychemia.utils.mathematics import lcm, shortest_triple_set
from pychemia import Structure
import itertools


class StructureMatch:
    def __init__(self, structure1, structure2):
        """
        Creates a structure match between 2 structures
        The structures will be change to match their number of
        atoms and the order of atoms inside such that the distances
        between atoms in equivalent positions on both structures
        is minimized.

        :param structure1: (Structure)
        :param structure2: (Structure)
        """
        assert (isinstance(structure1, Structure))
        assert (isinstance(structure2, Structure))
        assert structure1.is_perfect
        assert structure2.is_perfect
        self.structure1 = structure1.copy()
        self.structure2 = structure2.copy()
        self.base_lattice = self.structure1.lattice

    def match_size(self):

        assert self.structure1.is_crystal
        assert self.structure2.is_crystal

        gcd1 = self.structure1.get_composition().gcd
        gcd2 = self.structure2.get_composition().gcd

        sts = np.array(shortest_triple_set(lcm(gcd1, gcd2) / gcd1)).astype(int)
        supercell_multiples = sts[self.structure1.lattice.lengths.argsort()[::-1]]
        self.structure1 = self.structure1.supercell(supercell_multiples)

        sts = np.array(shortest_triple_set(lcm(gcd1, gcd2) / gcd2))
        supercell_multiples = sts[self.structure2.lattice.lengths.argsort()[::-1]]
        self.structure2 = self.structure2.supercell(supercell_multiples)

    def match_shape(self):

        self.structure1.canonical_form()
        self.structure2.canonical_form()
        assert (self.structure1.symbols == self.structure2.symbols)

    def match_atoms(self):

        if self.structure1.natom != self.structure2.natom:
            raise ValueError('Match the size first')

        best = {}
        for specie in self.structure1.species:
            selection = np.array(self.structure1.symbols) == specie

            distance_matrix, close_images = self.base_lattice.minimal_distances(self.structure1.reduced[selection],
                                                                                self.structure2.reduced[selection])

            min_trace = 1E10
            best[specie] = None
            if self.structure1.natom < 7:
                for i in itertools.permutations(range(len(distance_matrix))):
                    if distance_matrix[:, np.array(i)].trace() < min_trace:
                        min_trace = distance_matrix[:, np.array(i)].trace()
                        best[specie] = i
            else:
                # Only consider permutations of 2 positions
                if len(distance_matrix) > 1:
                    for ipar in itertools.permutations(range(len(distance_matrix)), 2):
                        i = list(range(len(distance_matrix)))
                        i[ipar[0]] = ipar[1]
                        i[ipar[1]] = ipar[0]
                        if distance_matrix[:, np.array(i)].trace() < min_trace:
                            min_trace = distance_matrix[:, np.array(i)].trace()
                            best[specie] = i
                    for ipar in itertools.permutations(range(len(distance_matrix)), 4):
                        i = list(range(len(distance_matrix)))
                        i[ipar[0]] = ipar[1]
                        i[ipar[1]] = ipar[0]
                        i[ipar[2]] = ipar[3]
                        i[ipar[3]] = ipar[2]
                        if distance_matrix[:, np.array(i)].trace() < min_trace:
                            min_trace = distance_matrix[:, np.array(i)].trace()
                            best[specie] = i
                else:
                    best[specie] = [0]

            print('For specie %s best permutation is %s' % (specie, str(best[specie])))

        best_permutation = np.zeros(self.structure1.natom, dtype=int)
        index = 0
        while index < self.structure1.natom:
            specie = self.structure1.symbols[index]
            selection = np.array(self.structure1.symbols) == specie
            best_permutation[selection] = index + np.array(best[specie])
            index += len(best[specie])

        self.structure2.sort_sites_using_list(best_permutation)

    def reduced_displacement(self):

        assert (self.structure1.symbols == self.structure2.symbols)
        assert (self.structure1.nsites == self.structure2.nsites)
        assert (self.structure1.natom == self.structure2.natom)
        ret = np.zeros((self.structure1.nsites, 3))

        distance_matrix, close_images = self.base_lattice.minimal_distances(self.structure1.reduced,
                                                                            self.structure2.reduced)

        for i in range(self.structure1.nsites):
            x1 = self.structure1.reduced[i]
            x2 = self.structure2.reduced[i] + close_images[i, i]
            ret[i] = x2 - x1

        return ret

    def cell_displacement(self):

        return np.dot(self.structure1.cell, np.linalg.inv(self.structure2.cell))

    def cartesian_distances(self):

        rd = self.reduced_displacement()
        ret = np.zeros(self.structure1.nsites)
        for i in range(self.structure1.nsites):
            ret[i] = np.dot(np.dot(rd[i], self.base_lattice.metric), rd[i])
            ret[i] = np.sqrt(ret[i])
        return ret



import os
from ..codes import CodeInput
from ...utils.computing import string2number


class ElkInput(CodeInput):

    def __init__(self, filename='elk.in'):

        CodeInput.__init__(self)
        if os.path.isfile(filename):
            self.input_file = filename
            self.read()

    def read(self):
        """
        Reads the input file and populate the variables dictionary
        """
        rf = open(self.input_file)
        lines = rf.readlines()
        rf.close()
        # All the variables stored as a python dictionary 
        ret = {}
        # Current key
        curkey = None
        index = -1
        for iline in range(len(lines)):
            # Remove leading and trailing spaces and newline jumps
            line = lines[iline].strip()
            # For debugging parser
            # print("%3d \t%s" % (iline,line))
            # Ignore empty lines
            if len(line) == 0:
                continue
            # Ignore lines that start with "!"
            elif line[0] == '!':
                continue
            elif index >= iline:
                continue
            # Special Parsing for atoms
            elif line == 'atoms':
                curkey = 'atoms'
                index = iline + 1
                line = lines[index]
                nspecies = int(line.split()[0])
                ret['atoms'] = {'order': [], 'nspecies': nspecies}
                for j in range(nspecies):
                    index += 1
                    line = lines[index].strip()
                    spfname = ''
                    for i in range(1, len(line)):
                        if line[i] == "'":
                            break
                        spfname += line[i]
                    ret['atoms']['order'].append(spfname)
                    index += 1
                    line = lines[index].strip()
                    natoms = int(line.split()[0])
                    ret['atoms'][spfname] = {'atposl': [], 'bfcmt': [], 'natoms': natoms}
                    for k in range(natoms):
                        index += 1
                        line = lines[index].strip()
                        atposl = [float(x) for x in line.split()[:3]]
                        ret['atoms'][spfname]['atposl'].append(atposl)
                        if len(line.split()) >= 6 and line.split()[4][0] != ':':
                            bfcmt = [float(x) for x in line.split()[3:6]]
                        else:
                            bfcmt = [0.0, 0.0, 0.0]
                        ret['atoms'][spfname]['bfcmt'].append(bfcmt)
            # When Line starts with an alpha character
            elif line[0].isalpha() or line == '.true.' or line == '.false.':
                # print("alph>%s" % line)
                if line == 'true' or line == '.true.':
                    ret[curkey].append(True)
                elif line == 'false' or line == '.false.':
                    ret[curkey].append(False)
                elif curkey is not None and curkey in ret:
                    if len(ret[curkey]) == 1:
                        ret[curkey] = ret[curkey][0]
                    curkey = line.split()[0]
                    ret[curkey] = []
                elif curkey is None:
                    curkey = line.split()[0]
                    ret[curkey] = []
                else:
                    print("ERROR: %s" % line)
            # When Line starts with a number
            elif line[0].isdigit() or (line[0] == '-' and line[1].isdigit()):
                # print("numb>%s" % line)
                for itoken in line.split():
                    if itoken[0] == ':':
                        break
                    elif itoken[0].isdigit() or (itoken[0] == '-' and itoken[1].isdigit()):
                        number, kind = string2number(itoken)
                        if number is None:
                            raise ValueError("Token could not be converted to number:%s" % itoken)
                        ret[curkey].append(number)
            # When line starts with single quotes
            elif line[0] == "'":
                # print("quot>%s" % line)
                word = ''
                for i in range(1, len(line)):
                    if line[i] == "'":
                        break
                    word += line[i]
                ret[curkey].append(word)
            else:
                print("Could not parse:%s" % line)
        self.variables = ret

    def __str__(self):
        """
        String representation of the input file.
        This string is used to show it to the user in interactive mode
        and to write it into a file with the method "write"

        """
        ret = ""
        for ikey in self.variables.keys():
            ret += '\n' + ikey + '\n'
            value = self.variables[ikey]
            if ikey == 'atoms':
                nspecies = value['nspecies']
                ret += " %d\n" % nspecies
                for i in range(nspecies):
                    spfname = value['order'][i]
                    ret += " '%s'\n" % spfname
                    natoms = value[spfname]['natoms']
                    ret += " %d\n" % natoms
                    for j in range(natoms):
                        ret += " %9.6f %9.6f %9.6f" % tuple(value[spfname]['atposl'][j])
                        ret += " %9.6f %9.6f %9.6f\n" % tuple(value[spfname]['bfcmt'][j])
            elif ikey == 'avec':
                for j in range(3):
                    ret += " %9.6f %9.6f %9.6f\n" % tuple(value[j * 3:j * 3 + 3])
            elif ikey == 'tasks':
                for itask in value:
                    ret += " %d\n" % itask
            elif ikey == 'wplot':
                ret += " %d %d %d\n" % tuple(value[:3])
                ret += " %9.3f %9.3f\n" % tuple(value[3:])
            else:
                if type(value) == list:
                    for i in value:
                        ret += " " + str(i)
                    ret += '\n'
                elif type(value) == int:
                    ret += " %d\n" % value
                elif type(value) == float:
                    ret += " %f\n" % value
                elif type(value) == str:
                    ret += " '%s'\n" % value
                elif type(value) == bool:
                    ret += "%s\n" % str(value).lower()
                else:
                    raise ValueError("Could not identify proper type for %s with type: %s" % (value, type(value)))

        return ret



"""
Structure contains the description of a set of atoms in space, for periodic structures it adds a lattice.
"""
try:
    import itertools.izip as zip
except ImportError:
    pass

import json
import os
import struct
import sys
import numpy as np
from collections.abc import MutableSequence
from itertools import combinations, repeat
from math import sin, cos
from multiprocessing import Pool
from pychemia import pcm_log
from pychemia.crystal.lattice import Lattice
from pychemia.core.composition import Composition
from pychemia.core.delaunay import get_reduced_bases
from pychemia.utils.computing import deep_unicode
from pychemia.utils.periodic import mass, atomic_number, covalent_radius, valence, atomic_symbols
import scipy.spatial


class Structure(MutableSequence):
    """
    Define an object that contains information about atomic positions,
    cell parameters and periodicity and provides methods to manipulate
    those elements

    A Structure is basically a set of sites with eventually a lattice
    each site could have one or more species with occupations equal
    or lower than one.

    A Structure could represents a molecule, cluster, wire, slab,
    crystal structure or alloy with define sites.
    The positions of the atoms and their atomic symbols are declared
    in 'positions' and 'symbols' respectively.

    For periodic structures, the 'periodicity' can be declared.
    and cell parameters in 'cell'

    Magnetic moments can be associated in the array vector_info['magnetic_moments'].
    """

    def __init__(self, natom=None, symbols=None, periodicity=False, cell=None, positions=None, reduced=None,
                 mag_moments=None, occupancies=None, sites=None, name=None, comment=None, vector_info=None):
        """ Structure is a container for geometric structure and composition for both periodic and non periodic
        atomic structures

        :param natom: Number of atoms
        :param symbols: List of atomic symbols
        :param periodicity: (True) if the structure is periodic, False for finite structures and a list of booleans for
        structures that are periodic only along specific directions.
        :param cell: The cell parameters, could be scalar for a cubic cell, a list of 3 numbers for a orthogonal cell
        or a complete numpy array or list of lists. Each row is considered a cell vector.
        :param positions: Array of rows with 3 elements for the positions of atoms in cartesian coordinates.
        :param reduced: Positions of atoms as reduced coordinates relative to the cell vectors ie cell-scaled in range
        [0,1]
        :param mag_moments: Magnetic moments for atoms in the structure
        :param occupancies: Atomic occupancies. 1 by default, lower values for vacancies and non-perfect crystals.
        :param sites: Atomic sites

        >>> a = Structure()
        >>> print(a)
        Empty structure
        >>> a = Structure(symbols=['Xe'])
        >>> print(a.natom)
        1
        >>> d = 1.104
        >>> a = Structure(symbols=['N', 'N'], positions=[[0, 0, 0], [0, 0, d]], periodicity=False)
        >>> print(a.natom)
        2
        >>> a = 4.05
        >>> b = a/2
        >>> fcc = Structure(symbols=['Au'], cell=[[0, b, b], [b, 0, b], [b, b, 0]], periodicity=True)
        >>> print(fcc.natom)
        1
        """
        self.vector_info = {}
        self.name = None
        self.comment = None
        self.natom = None
        self.symbols = None
        self.positions = None
        self.reduced = None
        self.cell = None
        self.periodicity = None
        self.vector_info['mag_moments'] = None
        self.sites = None
        self.occupancies = None
        self._lattice = None
        self._composition = None
        self.vector_info = None

        # By default the number of atoms will be the value given or zero except if other information overrules
        # that value
        if natom is not None:
            self.natom = int(natom)
        else:
            self.natom = 0

        if symbols is not None:
            if symbols in atomic_symbols:
                self.symbols = [symbols]
            else:
                for iatom in list(symbols):
                    assert(iatom in atomic_symbols)
                self.symbols = list(symbols)

            self.natom = len(self.symbols)
        else:
            if self.natom != 0:
                raise ValueError('List of atomic symbols not provided for structure with %d atoms', self.natom)

        # No periodicity will be assumed except if cell or reduced coordinates are provided
        if periodicity is None:
            periodicity = 3*[False]
        if isinstance(periodicity, bool):
            periodicity = 3*[periodicity]
        self.set_periodicity(periodicity)

        if cell is not None:
            cell = np.array(cell)
            self.set_cell(cell)
            self.periodicity = 3*[True]

        if positions is not None:
            positions = np.array(positions)
            self.set_positions(positions)
        if reduced is not None:
            reduced = np.array(reduced)
            self.set_reduced(reduced)
            self.periodicity = 3*[True]
        if mag_moments is not None:
            self.set_mag_moments(np.array(mag_moments))
        if occupancies is not None:
            self.occupancies = list(occupancies)
        if sites is not None:
            self.sites = sites

        if vector_info is None:
            self.vector_info = {'mag_moments': None}
        else:
            self.vector_info = vector_info

        self.name = name
        self.comment = comment

        # This routine completes the missing values and makes all the values coherent.
        self._autocomplete()

        if not self._check():
            raise ValueError('Arguments non consistent')

    def __len__(self):
        """ Number of sites in structure.
        In for perfect crystals it will match the number of atoms as each atomic site will have a single atom.

        :return: Number of sites in structure
        :rtype; int

        >>> st = Structure(symbols=['H', 'O'], positions= [[0,0,0], [0,0,1]])
        >>> len(st)
        2
        """
        return self.nsites

    def __str__(self):
        """String representation of Structure

        :return: Human readable text for Structure

        >>> st = Structure(symbols=['Xe'])
        >>> print(st)
        1
        <BLANKLINE>
        Symb  (             Positions            )
          Xe  (     0.0000     0.0000     0.0000 )
        <BLANKLINE>
        Non-periodic structure

        >>> a = 4.05
        >>> b = a/2
        >>> fcc = Structure(symbols=['Au'], cell=[[0, b, b], [b, 0, b], [b, b, 0]], periodicity=True)
        >>> print(fcc)
        1
        <BLANKLINE>
        Symb  (             Positions            ) [     Cell-reduced coordinates     ]
          Au  (     0.0000     0.0000     0.0000 ) [     0.0000     0.0000     0.0000 ]
        <BLANKLINE>
        Periodicity:  X Y Z
        <BLANKLINE>
        Lattice vectors:
             0.0000     2.0250     2.0250
             2.0250     0.0000     2.0250
             2.0250     2.0250     0.0000
        <BLANKLINE>
        """
        if self.natom == 0:
            xyz = 'Empty structure'
        else:
            xyz = str(self.natom) + '\n\n'
            if self.is_crystal:
                xyz += 'Symb  (             Positions            ) [     Cell-reduced coordinates     ]\n'
            else:
                xyz += 'Symb  (             Positions            )\n'

            for i in range(self.natom):
                if self.is_crystal:
                    xyz += ("%4s  ( %10.4f %10.4f %10.4f ) [ %10.4f %10.4f %10.4f ]\n"
                            % (self.symbols[i],
                               self.positions[i, 0],
                               self.positions[i, 1],
                               self.positions[i, 2],
                               self.reduced[i, 0],
                               self.reduced[i, 1],
                               self.reduced[i, 2]))
                else:
                    xyz += ("%4s  ( %10.4f %10.4f %10.4f )\n"
                            % (self.symbols[i],
                               self.positions[i, 0],
                               self.positions[i, 1],
                               self.positions[i, 2]))

            if self.periodicity[0] or self.periodicity[1] or self.periodicity[2]:
                xyz += '\nPeriodicity: '
                if self.periodicity[0]:
                    xyz += ' X'
                if self.periodicity[1]:
                    xyz += ' Y'
                if self.periodicity[2]:
                    xyz += ' Z'
                xyz += '\n\nLattice vectors:\n'
                for i in range(3):
                    xyz += (" %10.4f %10.4f %10.4f\n"
                            % (self.cell[i, 0], self.cell[i, 1], self.cell[i, 2]))
            else:
                xyz += '\nNon-periodic structure'

        return xyz

    def __repr__(self):
        """
        Evaluatable representation of Structure

        :return: String representation of the structure
        :rtype: str

        >>> st1 = Structure(symbols=['H'])
        >>> st2 = eval(repr(st1))
        >>> st1 == st2
        True
        >>> st = Structure(symbols='He', cell=[2,2,2])
        >>> st
        Structure(symbols=['He'], cell=2, reduced=[[0.0, 0.0, 0.0]], periodicity=True)
        """
        ret = 'Structure(symbols=' + str(self.symbols)
        if self.is_periodic:
            if np.all(np.diag(self.cell.diagonal()) == self.cell):
                if np.max(self.cell.diagonal()) == np.min(self.cell.diagonal()):
                    ret += ', cell=' + str(self.cell[0, 0])
                else:
                    ret += ', cell=' + str(self.cell.diagonal().tolist())
            else:
                ret += ', cell=' + str(self.cell.tolist())
            ret += ', reduced=' + str(self.reduced.tolist())
        else:
            ret += ', positions=' + str(self.positions.tolist())
        if all([self.periodicity[0] == item for item in self.periodicity]):
            ret += ', periodicity=' + str(self.periodicity[0])
        else:
            ret += ', periodicity=' + str(self.periodicity)
        ret += ')'
        return ret

    def __delitem__(self, key):
        self.del_atom(key)

    def __setitem__(self, key, value):
        self.add_atom(value['symbols'], value['positions'])

    def __getitem__(self, item):
        return SiteSet(self)[item]

    def __iter__(self):
        return iter(SiteSet(self))

    def insert(self, index, value):
        self.add_atom(value['symbols'], value['positions'])

    def _autocomplete(self):
        if self.natom is None:
            if self.positions is not None:
                self.natom = len(self.positions)
            elif self.reduced is not None:
                self.natom = len(self.reduced)
            elif self.symbols is not None:
                self.natom = len(self.symbols)
            else:
                self.natom = 0

        if self.symbols is None and self.natom == 0:
            self.symbols = []

        if self.periodicity is None:
            self.set_periodicity(True)

        if self.cell is None and self.is_periodic:
            self.set_cell(1)

        if self.positions is None:
            if self.reduced is not None:
                self.reduced2positions()
            else:
                if self.natom == 0:
                    self.positions = np.array([])
                elif self.natom == 1:
                    self.positions = np.array([[0.0, 0.0, 0.0]])
                else:
                    raise ValueError('Positions must be present for more than 1 atom')

        if self.reduced is None and self.is_crystal:
            if self.positions is not None and self.natom > 0:
                self.positions2reduced()
            else:
                self.reduced = np.array([])

        if self.sites is None:
            self.sites = range(self.natom)

        if self.occupancies is None:
            self.occupancies = self.natom * [1.0]

    def _check(self):
        check = True

        if len(self.symbols) != self.natom:
            print('Error: Bad symbols')
            check = False
        if len(self.positions) != self.natom:
            print('Error: Bad positions')
            check = False
        if self.is_crystal and len(self.reduced) != self.natom:
            print('Error: Bad reduced')
            check = False
        if self.vector_info['mag_moments'] is not None and len(self.vector_info['mag_moments']) != self.natom:
            print('Error: Bad mag_moments')
            check = False

        return check

    def add_atom(self, name, coordinates, option='cartesian'):
        """
        Add an atom with a given 'name' and cartesian or reduced 'position'
        The atom will be added at the end of the list of atoms in the Structure

        :param name: (str)
        :param coordinates: (list, numpy.array)
        :param option: (str)
        """
        assert (name in atomic_symbols)
        assert (option in ['cartesian', 'reduced'])
        self.symbols.append(name)
        self.natom += 1
        self._composition = None

        if option == 'cartesian':
            if self.natom == 0:
                self.positions = np.array(coordinates).reshape([-1, 3])
            else:
                self.positions = np.append(self.positions, coordinates).reshape([-1, 3])
            self.positions2reduced()
        elif option == 'reduced':
            if self.natom == 0:
                self.reduced = np.array(coordinates).reshape([-1, 3])
            else:
                self.reduced = np.append(self.reduced, coordinates).reshape([-1, 3])
            self.reduced2positions()

    def del_atom(self, index):
        """
        Removes the atom with the given index

        :param index:

        :return:
        """
        assert (abs(index) < self.natom)
        self.symbols.pop(index)
        np.delete(self.positions, index, 0)
        if self.is_periodic:
            np.delete(self.reduced, index, 0)
        self.natom -= 1
        self._composition = None

    def center_mass(self, list_of_atoms=None):
        """
        Computes the center of mass (CM) of the XYZ object or
        a partial list of atoms. The default is to compute the
        CM of all the atoms in the object, if a list
        is enter only those in the list will be included for the CM
        Return the CM as a numpy array
        """
        if list_of_atoms is None:
            list_of_atoms = range(self.natom)

        total_mass = 0.0
        center_of_mass = np.zeros(3)
        if self.natom == 0:
            return center_of_mass

        atomicnumber = atomic_number(list(self.symbols))

        for i in range(self.natom):
            if i in list_of_atoms:
                total_mass += mass(atomicnumber[i])
                center_of_mass += mass(atomicnumber[i]) * self.positions[i]

        return center_of_mass / total_mass

    def rotation(self, tx, ty, tz):
        """
        Rotate the molecule in the three directions
        """

        rotationx = np.array([[1, 0, 0], [0, cos(tx), -sin(tx)], [0, sin(tx), cos(tx)]])
        rotationy = np.array([[cos(ty), 0, sin(ty)], [0, 1, 0], [-sin(ty), 0, cos(ty)]])
        rotationz = np.array([[cos(tz), -sin(tz), 0], [sin(tz), cos(tz), 0], [0, 0, 1]])

        rotation = np.dot(np.dot(rotationx, rotationy), rotationz)

        for i in range(self.natom):
            self.positions[i] = np.dot(rotation, self.positions[i])

    def get_cell(self):
        if self._lattice is None:
            self._lattice = Lattice(self.cell)
        return self._lattice

    @property
    def lattice(self):
        return self.get_cell()

    def get_composition(self, gcd=True):
        """
        Computes the composition of the Structure
        as the count of each species in the cell
        If gcd is True the values are divided by the
        greatest common divisor

        :param gcd: bool

        :rtype : Composition
        """
        if self._composition is None:
            species = {}
            for atom in self.symbols:
                if atom in species:
                    species[atom] += 1
                else:
                    species[atom] = 1
            self._composition = Composition(species)
        return self._composition

    def positions2reduced(self):
        """
        Computes the cell-reduced coordinates from the
        cartesian dimensional coordinates
        """
        self.reduced = np.linalg.solve(self.cell.T, self.positions.T).T
        for i in range(3):
            if self.periodicity[i]:
                self.reduced[:, i] %= 1.0

    def reduced2positions(self):
        """
        Computes the dimensional cartesian coordinates
        from the adimensional cell-reduced coordinates
        """
        self.positions = np.dot(self.reduced, self.cell)

    def relocate_to_cm(self, list_of_atoms=None):
        """
        Relocates the system of atoms to the center of mass
        a partial list of atoms can be used to compute
        the center, but all the atoms are moved to the
        computed center

        :param list_of_atoms: (list) List of atoms that will be considered for computing the center of mass
                              by default all atoms are included
        """
        cm = self.center_mass(list_of_atoms)
        self.positions += -cm

    def get_distance(self, iatom, jatom, with_periodicity=True, tolerance=1e-5):
        """
        Calculates the distance between 2 atom, identified by index
        iatom and jatom

        :param iatom: (int) index of first atom
        :param jatom: (int) index of second atom
        :param with_periodicity: (bool) if the periodic images should be considered to compute the shortest distance
        :param tolerance: (float) Tolerance for the bases reduction

        :rtype : (float) distance between iatom and jatom
        """

        if with_periodicity:
            reduced_bases = get_reduced_bases(self.cell, tolerance)
            scaled_pos = np.dot(self.positions, np.linalg.inv(reduced_bases))
            # move scaled atomic positions into -0.5 < r <= 0.5
            for pos in scaled_pos:
                pos -= pos.round()

            # Look for the shortest one in surrounded 3x3x3 cells
            distances_list = []
            for i in (-1, 0, 1):
                for j in (-1, 0, 1):
                    for k in (-1, 0, 1):
                        distances_list.append(np.linalg.norm(
                            np.dot(scaled_pos[iatom] - scaled_pos[jatom] +
                                   np.array([i, j, k]), reduced_bases)))
            ret = min(distances_list)

        else:
            posi = self.positions[iatom]
            posj = self.positions[jatom]
            ret = np.linalg.norm(posi - posj)

        return ret

    @staticmethod
    def random_cell(composition, method='stretching', stabilization_number=20, nparal=5, periodic=True,
                    factor_optimal_volume=8):
        """
        Generate a random cell
        There are two algorithms implemented:

        scaling: Generate a random cell and random distribution of atoms and
                scale the lattice to separate the atoms.

        stretching: Generating a random cell and random distribution of atoms
                    and stretching their bonds until the distance between any
                    two atoms is always greater than the sum of covalent radius.

        :param composition: (pychemia.Composition)
        :param method: (str)
        :param stabilization_number: (int)
        :param nparal: (int)
        :param periodic: (bool)
        :param factor_optimal_volume: (float)
        :return:

        >>> import os
        >>> st = Structure.random_cell('LiAlCl4', stabilization_number=3)
        >>> st.natom
        6
        >>> st.save_json('test.json')
        >>> st2 = Structure.load_json('test.json')
        >>> st == st2
        True
        >>> os.remove('test.json')
        """
        comp = Composition(composition)
        pcm_log.debug('Generating a random structure with composition: ' + str(comp.composition))
        natom = comp.natom
        symbols = comp.symbols

        best_volume = sys.float_info.max
        best_volume = float('inf')
        best_structure = None
        optimal_volume = comp.covalent_volume('cubes')
        stabilization_history = 0
        pool = Pool(processes=nparal)

        trial = 0
        while stabilization_history < stabilization_number:
            args = list(best_volume * np.ones(10))

            ret = pool.map(worker_star, zip(repeat(method), repeat(composition), repeat(periodic), args))

            ngood = 0
            for structure in ret:
                if structure is not None:
                    # print('SH:%d Vol:%10.3f Factor:%10.3f' % (stabilization_history,
                    #                                          structure.volume,
                    #                                          structure.volume / optimal_volume))
                    ngood += 1
                    if best_structure is not None:
                        if structure.volume < best_structure.volume:
                            best_structure = structure
                    else:
                        best_structure = structure

            # log.debug('Good structures: %d/10  Best volume: %7.3f' % (ngood, best_structure.volume))
            if best_structure is not None and best_volume > best_structure.volume:
                best_volume = best_structure.volume
                stabilization_history = 0
            else:
                stabilization_history += 1

            if best_volume < factor_optimal_volume * optimal_volume:
                break
            trial += 1

            # log.debug('Trial: %4d  Volume: %7.2f  Optimal Volume: %7.2f  Ratio: %5.2f' %
            #          (trial, best_volume, optimal_volume, best_volume/optimal_volume))

        pool.close()

        if best_structure is not None and periodic:
            # Analysis of the quality for the best structure
            rpos = best_structure.reduced
            for i, j in combinations(range(natom), 2):
                distance = best_structure.lattice.minimal_distance(rpos[i], rpos[j])
                covalent_distance = sum(covalent_radius([symbols[i], symbols[j]]))
                if distance < covalent_distance:
                    pcm_log.debug('Covalent distance: %7.4f  Minimal distance: %7.4f  Difference: %7.3e' %
                                  (covalent_distance, distance, covalent_distance - distance))

        best_structure.canonical_form()
        return best_structure

    @staticmethod
    def random_cluster(composition, method='stretching', stabilization_number=20, nparal=5):

        st = Structure.random_cell(composition=composition, method=method, stabilization_number=stabilization_number,
                                   nparal=nparal, periodic=False)
        return Structure(symbols=st.symbols, positions=st.positions, periodicity=False)

    def adjust_reduced(self):
        for i in range(self.natom):
            for j in range(3):
                for value in [0.5, 0.25, 0.75, 0.125]:
                    if abs(value - self.reduced[i, j]) < 1E-4:
                        self.reduced[i, j] = value
        self.reduced2positions()

    def set_cell(self, cell):
        """
        Set the vectors defining the cell

        :param cell: A matrix with the 3 unit cell
                     vectors
        :return:
        """
        npcell = np.array(cell)
        if npcell.shape == () or npcell.shape == (1,):
            self.cell = npcell * np.eye(3)
        elif npcell.shape == (3,):
            self.cell = np.diag(npcell)
        else:
            self.cell = np.array(cell).reshape((3, 3))
        self._lattice = None

    def set_mag_moments(self, mag_moments):
        """
        Set the magnetic moments with one vector on each
        atom

        Args:
            mag_moments: List or numpy array with one
            vector for each atom.
            The values will be converted into a numpy array
        """
        self.vector_info['mag_moments'] = np.array(mag_moments).reshape([-1, 3])

    def set_periodicity(self, periodicity):
        """
        Set periodicity of the structure

        Args:
            periodicity: (Boolean) a single value means that the structure has that
                        periodicity all along the 3 directions. Otherwise a list
                        of 3 booleans is required
        """
        if isinstance(periodicity, bool):
            self.periodicity = 3 * [periodicity]
        elif isinstance(periodicity, list) and len(periodicity) == 1:
            self.periodicity = 3 * periodicity
        else:
            self.periodicity = list(periodicity)

    def set_positions(self, positions):
        """
        Set the positions of the atoms
        This contains dimensional values
        in cartesian coordinates

        Args:
            positions: A array of 3 vectors
            with dimensional coordinates
        """
        self.positions = np.array(positions).reshape([-1, 3])

    def set_reduced(self, reduced):
        """
        Set the reduced positions of the atoms
        This contains adimensional values
        relative to cell vectors

        :param reduced:
        :return:
        """
        self.reduced = np.array(reduced).reshape([-1, 3])

    def sort_sites_using_list(self, sorted_indices):
        sorted_indices = np.array([int(x) for x in sorted_indices])
        self.symbols = list(np.array(self.symbols)[sorted_indices])
        self.positions = self.positions[sorted_indices]
        if self.is_periodic:
            self.reduced = self.reduced[sorted_indices]
        if self.vector_info is not None:
            for vi in self.vector_info:
                if self.vector_info[vi] is not None:
                    self.vector_info[vi] = self.vector_info[vi][sorted_indices]

    def sort_sites(self):

        # First: Sort sites using the distance to the origin
        sorted_indices = np.array([np.linalg.norm(self.positions[i]) for i in range(self.nsites)]).argsort()
        # print sorted_indices
        self.sort_sites_using_list(sorted_indices)

        # Second: Sort again using the atomic number
        if len(self.species) > 1:
            sorted_indices = np.array([atomic_number(x) for x in self.symbols]).argsort()
            self.sort_sites_using_list(sorted_indices)

    def sort_axes(self):
        """
        Sort the lattice vectors in decremental order of their size.
        'a' will be the longest lattice vector
        'c' he shortest
        """

        sorted_indices = self.lattice.lengths.argsort()[::-1]
        self.set_cell(self.cell[sorted_indices])
        self.reduced = self.reduced[:, sorted_indices]

    def align_with_axis(self, axis=0, round_decimals=14):
        lattice = self.lattice
        lattice.align_with_axis(axis=axis, round_decimals=round_decimals)
        self.set_cell(lattice.cell)
        self.reduced2positions()

    def align_with_plane(self, axis=2, round_decimals=14):
        lattice = self.lattice
        lattice.align_with_plane(axis=axis, round_decimals=round_decimals)
        self.set_cell(lattice.cell)
        self.reduced2positions()

    def align_inertia_momenta(self):
        ii = self.inertia_matrix()
        eigval, eigvec = np.linalg.eig(ii)
        eigvec = eigvec.T[eigval.argsort()[::-1]].T
        inveigvec = np.linalg.inv(eigvec)
        self.positions = np.dot(inveigvec, self.positions.T).T

    def canonical_form(self):

        if not self.is_periodic:
            self.relocate_to_cm()
            self.align_inertia_momenta()
        self.sort_sites()
        if self.is_periodic:
            self.sort_axes()
            self.align_with_axis()
            self.align_with_plane()
            self.atoms_in_box()
            self.sort_sites()

    def supercell(self, size):
        """
        Creates a supercell, replicating the positions
        of atoms in the x,y,z directions a number of
        size=(nx,ny,nz) times
        """
        new_natom = np.prod(size) * self.natom
        new_symbols = []
        new_positions = np.zeros((new_natom, 3))
        size = np.array(size).astype(int)

        index = 0
        for i in range(size[0]):
            for j in range(size[1]):
                for k in range(size[2]):
                    for n in range(self.natom):
                        new_symbols.append(self.symbols[n])
                        new_positions[index] = self.positions[n] + (
                            i * self.cell[0] + j * self.cell[1] + k * self.cell[2])
                        index += 1
        new_cell = np.zeros((3, 3))
        new_cell[0] = size[0] * self.cell[0]
        new_cell[1] = size[1] * self.cell[1]
        new_cell[2] = size[2] * self.cell[2]
        return Structure(symbols=new_symbols, positions=new_positions, cell=new_cell)

    def copy(self):
        """
        Get a copy of the object
        """
        copy_struct = Structure(name=self.name, comment=self.comment, natom=self.natom, symbols=self.symbols,
                                periodicity=self.periodicity, cell=self.cell, positions=self.positions,
                                reduced=self.reduced, vector_info=self.vector_info, sites=self.sites,
                                occupancies=self.occupancies)
        return copy_struct

    @property
    def to_dict(self):

        ret = {'natom': self.natom,
               'symbols': self.symbols,
               'periodicity': self.periodicity,
               'positions': self.positions.tolist(),
               'nspecies': len(self.species),
               'formula': self.formula}
        if self.is_periodic:
            ret['cell'] = self.cell.tolist()
            ret['reduced'] = self.reduced.tolist()
            ret['density'] = self.density
        if self.name is not None:
            ret['name'] = self.name
        if self.comment is not None:
            ret['comment'] = self.comment
        if self.sites != range(self.natom):
            ret['sites'] = list(self.sites)
        if self.occupancies != self.natom * [1.0]:
            ret['occupancies'] = self.occupancies
        # if len(self.vector_info) != 1 or self.vector_info['mag_moments'] is not None:
        #    ret['vector_info'] = self.vector_info
        return ret

    def round(self, decimals=6, pos='reduced'):
        self.set_cell(np.around(self.cell, decimals))
        if pos == 'reduced':
            self.set_reduced(np.around(self.reduced, decimals))
            self.reduced2positions()
        else:
            self.set_positions(np.around(self.positions, decimals))
            self.positions2reduced()

    @staticmethod
    def from_dict(structdict):

        natom = structdict['natom']
        symbols = deep_unicode(structdict['symbols'])
        periodicity = structdict['periodicity']
        positions = np.array(structdict['positions'])

        if 'name' in structdict:
            name = structdict['name']
        else:
            name = None
        if 'comment' in structdict:
            comment = structdict['comment']
        else:
            comment = None
        if 'cell' in structdict:
            cell = np.array(structdict['cell'])
        else:
            cell = None
        if 'reduced' in structdict:
            reduced = np.array(structdict['reduced'])
        else:
            reduced = None
        if 'vector_info' in structdict:
            vector_info = structdict['vector_info']
        else:
            vector_info = None
        if 'sites' in structdict:
            sites = structdict['sites']
        else:
            sites = range(natom)
        if 'occupancies' in structdict:
            occupancies = structdict['occupancies']
        else:
            occupancies = list(np.ones(natom))
        return Structure(name=name, comment=comment, natom=natom, symbols=symbols, periodicity=periodicity, cell=cell,
                         positions=positions, reduced=reduced, vector_info=vector_info, sites=sites,
                         occupancies=occupancies)

    def save_json(self, filename):

        filep = open(filename, 'w')
        json.dump(self.to_dict, filep, sort_keys=True, indent=4, separators=(',', ': '))
        filep.close()

    @staticmethod
    def load_json(filename):

        filep = open(filename, 'r')
        structdict = deep_unicode(json.load(filep))
        filep.close()
        return Structure.from_dict(structdict)

    def distance2(self, atom1, atom2):
        assert (isinstance(atom1, int))
        assert (isinstance(atom2, int))
        assert (atom1 < self.natom)
        assert (atom2 < self.natom)

        if self.is_periodic:
            return self.lattice.distance2(self.reduced[atom1], self.reduced[atom2])
        else:
            dm = scipy.spatial.distance_matrix(self.positions, self.positions)
            return dm[atom1, atom2]

    def distance_matrix(self):
        if self.is_periodic:
            dm = np.zeros((self.nsites, self.nsites))
            for i in range(self.nsites - 1):
                for j in range(i + 1, self.nsites):
                    d = self.lattice.distance2(self.reduced[i], self.reduced[j], radius=1E10, limits=[1, 1, 1])
                    # print("%d %d - %d" % (i,j, len(d)))
                    dm[i, j] = min([d[x]['distance'] for x in d])
                    dm[j, i] = dm[i, j]
        else:
            dm = scipy.spatial.distance_matrix(self.positions, self.positions)
        return dm

    def valence_electrons(self):
        ret = 0
        for key, value in self.composition.items():
            ret += value * valence(key)
        return ret

    def __eq__(self, other):
        if self.natom != other.natom:
            ret = False
        elif not np.array_equal(self.positions, other.positions):
            ret = False
        elif not np.array_equal(self.periodicity, other.periodicity):
            ret = False
        elif self.is_periodic and other.is_periodic:
            if not np.array_equal(self.reduced, other.reduced):
                ret = False
            elif not np.array_equal(self.cell, other.cell):
                ret = False
            else:
                ret = True
        else:
            ret = True
        return ret

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def is_perfect(self):
        """
        Return True if two conditions are met:

        1. The number of sites is equal to
        the number of atoms. ie there is no more than
        one atom on each site.

        2. All the occupancies are equal to 1

        :rtype : bool

        :return: bool
        """
        return self.natom == self.nsites and min(self.occupancies) == 1.0

    @property
    def is_periodic(self):
        """
        Return True if the Structure is periodic in any direction
        False for non-periodic structures

        :rtype : bool

        :return: bool
        """
        return any(self.periodicity)

    @property
    def is_crystal(self):
        """
        True if structure is periodic in all directions
        False otherwise

        :rtype : bool

        :return: bool
        """
        if not self.is_periodic:
            return False
        else:
            return self.get_cell().periodic_dimensions == 3

    @property
    def composition(self):
        """
        Dictionary with the composition, the keys are the species and the values
        represent the number of atoms of that specie

        :rtype : dict

        :return: dict
        """
        return self.get_composition().composition

    @property
    def formula(self):
        """
        String with the chemical formula

        :rtype: str

        :return: str
        """
        return self.get_composition().formula

    @property
    def density(self):
        """
        Computes the density of the cell

        :rtype: float

        :return: float
        """
        return sum(np.array(mass(self.symbols))) / self.volume

    @property
    def volume(self):
        """
        Computes the volume of the cell

        :rtype: float

        :return: float
        """
        if self.is_periodic:
            return abs(np.linalg.det(self.cell))
        else:
            volume = (np.max(self.positions[:, 0]) - np.min(self.positions[:, 0])) * \
                     (np.max(self.positions[:, 1]) - np.min(self.positions[:, 1])) * \
                     (np.max(self.positions[:, 2]) - np.min(self.positions[:, 2]))

            if volume > 0.0:
                return volume
            else:
                return 4.0 / 3.0 * np.pi * np.max(self.positions.flatten()) ** 3

    @property
    def species(self):
        return self.get_composition().species

    @property
    def nspecies(self):
        return len(self.get_composition().species)

    @property
    def nsites(self):
        return len(self.positions)

    def scale(self, tolerance=0.7):
        assert self.is_perfect
        assert self.is_crystal
        lattice = self.lattice.scale(self.symbols, self.reduced, tolerance=tolerance)
        return Structure(cell=lattice.cell, reduced=self.reduced, symbols=self.symbols)

    # Not working
    # def cut_void(self, factor=1.5):
    #     ret = self.copy()
    #     ret.canonical_form()
    #     mins = [min(ret.reduced[:, i]) for i in range(3)]
    #     ret.reduced = ret.reduced - mins
    #     ret.reduced2positions()
    #     max_lenght = max(ret.positions[:, 0]) + 2.0*max([covalent_radius(ret.symbols[i]) for i in range(ret.nsites)])
    #     print factor, max_lenght, ret.cell[0,0]
    #     if factor * max_lenght < ret.cell[0, 0]:
    #         cell = ret.cell
    #         cell[0, 0] = factor * max_lenght
    #         ret.set_cell(cell)
    #         return ret
    #     else:
    #         return self

    def atoms_in_box(self):
        while min(self.reduced.flatten()) < 0.0 or max(self.reduced.flatten()) > 1.0:
            self.reduced = (self.reduced + 1.0) % 1.0
            self.reduced2positions()

    def moment_of_inertia(self, axis):

        assert self.is_perfect
        mofi = 0
        for isite in self:
            mofi += mass(isite.symbols[0]) * (sum(np.array(isite.position) ** 2) - isite.position[axis] ** 2)
        return mofi

    def product_of_inertia(self, axis):

        assert self.is_perfect
        pofi = 0
        for isite in self:
            pofi += mass(isite.symbols[0]) * (np.prod(isite.position) / isite.position[axis])
        return pofi

    def inertia_matrix(self):

        im_xx = self.moment_of_inertia(0)
        im_yy = self.moment_of_inertia(1)
        im_zz = self.moment_of_inertia(2)
        im_xy = self.product_of_inertia(2)
        im_xz = self.product_of_inertia(1)
        im_yz = self.product_of_inertia(0)

        im = np.array([[im_xx, -im_xy, -im_xz], [-im_xy, im_yy, -im_yz], [-im_xz, -im_yz, im_zz]])
        return im

    def signature(self):
        comp = self.get_composition()
        gcd = self.get_composition().gcd
        ret = '%02X_%014X_%02X_' % (self.valence_electrons() / gcd, comp.species_hex(), gcd)

        formula = "%s" % comp.sorted_formula(sortby='electroneg')
        formula += (17 - len(formula)) * '_'
        ret += formula

        return ret

    def add_vacuum(self, length, direction=2):

        vacuum = np.zeros(3)
        vacuum[direction] = length
        alpha = self.lattice.alpha
        beta = self.lattice.beta
        gamma = self.lattice.gamma
        newlenghts = self.lattice.lengths + vacuum
        a = newlenghts[0]
        b = newlenghts[1]
        c = newlenghts[2]
        newlattice = self.lattice.from_parameters_to_cell(a, b, c, alpha, beta, gamma)
        return Structure(symbols=self.symbols, cell=newlattice.cell, positions=self.positions)


def load_structure_json(filename):
    ret = Structure()
    ret.load_json(filename)
    return ret


class SiteSet:
    """
    Collection of atomic sites.
    Starting from a Structure the object will create a set of Sites storing the species, occupancies and
    positions of each site.
    """
    def __init__(self, structure):
        """
        SiteSet is a container for a list of Sites

        :param structure: structure from which the Sites will be created.
        """
        self.structure = structure
        self.sitelist = []
        reduced = None

        for isite in range(structure.nsites):
            if structure.sites.count(isite) > 1:
                symbols = []
                occupancies = []
                for jatom in range(structure.natom):
                    if structure.sites[jatom] == isite:
                        symbols.append(structure.symbols[jatom])
                        occupancies.append(structure.occupancies[jatom])
                position = structure.positions[isite]
                if self.structure.is_periodic:
                    reduced = structure.reduced[isite]
            else:
                symbols = [structure.symbols[isite]]
                occupancies = [structure.occupancies[isite]]
                position = structure.positions[isite]
                if self.structure.is_periodic:
                    reduced = structure.reduced[isite]
            self.sitelist.append(Site(symbols=symbols, occupancies=occupancies, position=position))

    def __iter__(self):
        return iter(self.sitelist)


class Site:
    """
    A site is a mapping of one location of space with one or more atoms and the corresponding occupancies.
    The sum of occupancies must be lower or equal to 1.
    """
    def __init__(self, symbols, occupancies=None, position=None):
        """
        Create a atomic site with one or more atoms defined on a unique location on space.

        :param symbols: atomic symbol or list of atomic symbols
        :param occupancies: list of floats with the probability of occupancy for the atom in the site
        :param position: iterable with the location of the atom in space in cartesian coordinates

        .. notes: This class is used for internal operations inside Structure. Not much reason to expose it
        out of this module.

        >>> sit = Site('H')
        >>> sit
        Site(symbols=['H'],occupancies=[1.0],position=[0.0, 0.0, 0.0])
        >>> sit = Site(symbols = ['O', 'N'], occupancies=[0.5, 0.5], position=[1,1,1])
        >>> sit
        Site(symbols=['O', 'N'],occupancies=[0.5, 0.5],position=[1.0, 1.0, 1.0])
        """
        if symbols in atomic_symbols:
            self.symbols = [symbols]
        else:
            try:
                self.symbols = list(symbols)
            except TypeError:
                raise TypeError("%s is not iterable" % symbols)
        for i in self.symbols:
            assert i in atomic_symbols

        if occupancies is None:
            occupancies = 1.0
        if position is None:
            position = 3*[0.0]

        try:
            self.occupancies = [float(x) for x in occupancies]
        except TypeError:
            self.occupancies = [float(occupancies)]

        assert (len(self.occupancies) == len(self.symbols))
        assert sum(self.occupancies) <= 1

        self.position = [float(x) for x in position]

    def __repr__(self):
        ret = self.__class__.__name__+'(symbols=' + repr(self.symbols)
        ret += ',occupancies=' + repr(self.occupancies)
        ret += ',position=' + repr(self.position)
        ret += ')'
        return ret

    def __str__(self):
        return repr(self)


def worker_star(x):
    return random_structure(*x)


def random_structure(method, composition, periodic=True, max_volume=1E10):
    """
    Random Structure created  by random positioning of atoms followed by either scaling of the cell or
    adding a sheer stretching along the smaller distances. The purpose of the lattice change is to avoid any two
    atoms to be closer than the sum of their covalent radius.

    :param method: Can be 'stretching' or 'scaling'.
    :param composition: Can be a Composition object or formula.
    :param periodic: If True, the structure will be periodical in all directions, otherwise a finite system is created.
    :param max_volume: Threshold for creating the Structure, if the volume exceeds the target the method returns None
    :return: Structure if the volume is below than best_volume, None otherwise

    >>> st = random_structure(method='scaling', composition='H2O', periodic=False)
    >>> st.natom
    3
    >>> st = random_structure(method='stretching', composition='NaCl', periodic=True)
    >>> st.natom
    2
    """
    comp = Composition(composition)
    natom = comp.natom
    symbols = comp.symbols
    np.random.seed(struct.unpack("<L", os.urandom(4))[0])

    if periodic:
        new_structure = None

        assert (method in ['scaling', 'stretching'])

	
        while True:
            trial=0
    
            if method == 'scaling':
                lattice = Lattice.random_cell(comp)
                # Random reduced positions
                rpos = np.random.rand(natom, 3)
                mins = [min(rpos[:, i]) for i in range(3)]
                rpos -= mins

                new_lattice = lattice
            else:
                lattice = Lattice.random_cell(comp)
                # Random reduced positions
                rpos = np.random.rand(natom, 3)
                mins = [min(rpos[:, i]) for i in range(3)]
                rpos -= mins

                new_lattice = lattice.stretch(symbols, rpos, tolerance=1.0, extra=0.1)

            if new_lattice.volume < max_volume:
                test = True
                for i in range(natom):
                    for j in range(i + 1, natom):
                        distance = new_lattice.minimal_distance(rpos[i], rpos[j])
                        covalent_dim = sum(covalent_radius([symbols[i], symbols[j]]))
                        if distance < covalent_dim:
                            test = False
                if test:
                    new_structure = Structure(symbols=symbols, reduced=rpos, cell=new_lattice.cell, periodicity=True)
                    minimal_distance = np.min(new_structure.distance_matrix() + 10 * np.eye(new_structure.natom))
                    break
                else:
                    print("Trial failed, distance %f is less than covalent radious %f" % (distance, covalent_dim))
                    trial += 1

            if trial > 100:
                print("Leaving after 100 attemps")
                break

        # else:
        #     print('Volume of Structure %f is larger than max_volume=%f' % (new_lattice.volume, max_volume))
        #     new_structure = None
    else:
        pos = np.random.rand(natom, 3)

        mindis = cluster_minimal_distance(pos)
        if mindis == 0:
            raise ValueError("Distance too small")

        max_cov = np.max(covalent_radius(symbols))
        pos *= max_cov / mindis

        current_volume = (max(pos[:, 0]) - min(pos[:, 0])) * (max(pos[:, 1]) - min(pos[:, 1])) * (
            max(pos[:, 2]) - min(pos[:, 2]))

        if current_volume < max_volume:
            new_structure = Structure(symbols=symbols, positions=pos, periodicity=False)
        else:
            print('Volume of Structure %f is larger than max_volume=%f' % (current_volume, max_volume))
            new_structure = None

    return new_structure
    # End of Worker


def cluster_minimal_distance(pos):
    pos = np.array(pos).reshape((-1, 3))
    dismat = scipy.spatial.distance_matrix(pos, pos)
    tmp = np.max(dismat.flatten())
    return np.min((dismat + tmp * np.eye(len(pos))).flatten())



"""
Chemical composition is just the description of the amount of atoms of each specie. In the case of clusters or
molecules, ie a finite structure, it represents the complete set of atoms. For periodic structures it represents
the species present on a cell.
"""

import re
from numpy import array, argsort
from math import gcd as _gcd
from math import pi
from pychemia.utils.periodic import atomic_symbols, electronegativity, atomic_number, covalent_radius
from pychemia.utils.computing import deep_unicode
from functools import reduce
from collections.abc import Mapping


class Composition(Mapping):
    """
    A Composition is basically a mapping between a number of species and a integer indicating how many atoms of that
    specie are present in the structure.
    A composition object do not contain geometrical information or bonding.

    The main purpose of this class is to be able to parse formulas into compositions and return string formulas sorted
    in various ways.
    """

    def __init__(self, value=None):
        """
        Creates a new composition,  currently only absolute formulas are supported.

        :param value: (str, dict) The input argument could be a string with a chemical formula or the actual dictionary
        of species and values. The order of species is not guaranteed to be preserved. A iterable of atomic symbols
        is also accepted to build a composition object.

        :rtype: Composition

        >>> cp = Composition({'Ba': 2, 'Cu': 3, 'O': 7, 'Y': 1})
        >>> cp.formula
        'Ba2Cu3O7Y'
        >>> cp = Composition('Ba2Cu3O7Y')
        >>> cp2 = Composition(cp)
        >>> len(cp2)
        4
        >>> cp.nspecies
        4
        >>> cp = Composition(['O', 'H', 'O'])
        >>> len(cp)
        2
        >>> cp['O']
        2

        """
        # The internal dictionary where atom species and numbers of atoms of each specie are stored.
        self._composition = {}
        # Convert strings and dictionaries into unicode
        if value is not None:
            value = deep_unicode(value)
        # Case 1: The input is a formula
        if isinstance(value, str):
            self._set_composition(self.formula_parser(value))
        # Case 2: The input is a dictionary
        elif isinstance(value, dict):
            self._set_composition(value)
        # Case 3: The input is another composition object
        elif isinstance(value, Composition):
            self._set_composition(value.composition)
        # Case 4: The input is an iterable of atomic symbols
        elif hasattr(value, "__len__"):
            dvalue = {}
            for i in value:
                if i in dvalue:
                    dvalue[i] += 1
                else:
                    dvalue[i] = 1
            self._set_composition(dvalue)
        else:
            self._composition = {}

    def __len__(self):
        return len(self._composition)

    def __getitem__(self, specie):
        """
        Returns the number of atoms of a given specie

        :param specie: Atomic Symbol for which the value will be returned
        :return: number of atoms of the given specie
        :rtype: int

        >>> comp = Composition('H2')
        >>> comp['H']
        2
        >>> comp['He']
        0
        """
        if specie in self._composition:
            return self._composition[specie]
        else:
            return 0

    def __repr__(self):
        """
        Evaluable representation of Composition object

        :return: Text representation that can be evaluated
        :rtype: str

        >>> cp1 = Composition('H2O')
        >>> cp2 = eval(repr(cp1))
        >>> cp2 == cp1
        True
        """
        return 'Composition(' + str(self.composition) + ')'

    def __str__(self):
        """

        :return: String representation of the composition

        >>> cp = Composition('YBa2Cu3O7')
        >>> 'Cu' in str(cp)
        True
        """
        ret = ''
        for i in self.species:
            ret += " %3s: %4d  " % (i, self.composition[i])
        return ret

    def __iter__(self):
        return iter(self.composition)

    def __contains__(self, specie):
        """True if 'specie' is present in composition

        :return: True if specie is present
        :param  specie: atomic specie
        :rtype: bool

        >>> cp = Composition('H2O')
        >>> 'He' in cp
        False
        """
        return specie in self._composition

    def _set_composition(self, value):
        """
        Checks the values of a dictionary before setting the actual composition

        :param value: (dict)
        :rtype: None
        """
        for i in value:
            assert (i in atomic_symbols)
            assert (isinstance(value[i], int))
        self._composition = value.copy()

    @property
    def composition(self):
        """Dictionary with composition

        :return: The composition dictionary
        :rtype: dict

        >>> import pprint
        >>> cp = Composition('H2O')
        >>> pprint.pprint(cp.composition)
        {'H': 2, 'O': 1}

        """
        return self._composition

    def covalent_volume(self, packing='cubes'):
        """
        :param packing: The kind of packing could be 'cubes' or 'spheres'
        :type packing: str
        :return: The volume occupied by a given formula assuming a 'cubes' packing or 'spheres' packing
        :rtype: (float)

        >>> cp = Composition('C5H10')
        >>> cp.covalent_volume()
        19.942320000000002
        >>> cp.covalent_volume(packing='spheres')
        10.441774334589468
        """
        if packing == 'cubes':
            factor = 8
        elif packing == 'spheres':
            factor = 4 * pi / 3.0
        else:
            raise ValueError('Non-valid packing: "%s"' % packing)

        # find volume of unit cell by adding cubes
        volume = 0.0
        for specie in self:
            number_atoms_specie = self.composition[specie]
            # Pack each atom in a cube (2*r)^3
            volume += factor * number_atoms_specie * covalent_radius(specie) ** 3
        return volume

    @property
    def formula(self):
        """Chemical formula

        :return: The chemical formula with atoms sorted alphabetically
        :rtype: str

        >>> cp = Composition('NaCl')
        >>> cp.formula
        'ClNa'

        """
        return self.sorted_formula(sortby='alpha', reduced=True)

    @staticmethod
    def formula_parser(value):
        """Return a dictionary from a chemical formula

        :return: Convert an string representing a chemical formula into a dictionary with the species as keys
                 and values as the number of atoms of that specie
        :param value: (str) Chemical formula
        :rtype: dict

        >>> import pprint
        >>> Composition.formula_parser('Au20')
        {'Au': 20}
        >>> ret = Composition.formula_parser('UutUupUusUuo')
        >>> pprint.pprint(ret)
        {'Uuo': 1, 'Uup': 1, 'Uus': 1, 'Uut': 1}
        """
        ret = {}
        jump = False
        for i in range(len(value)):
            if jump > 0:  # This char belongs to the current atom, move on
                jump -= 1
            elif value[i].isupper():  # Atom Name starts with Uppercase
                if i + 1 < len(value) and value[i + 1].islower():  # Atom name has more than 1 char
                    if i + 2 < len(value) and value[i + 2].islower():  # Atom name has more than 2 chars
                        specie = value[i:i + 3]
                        jump = 2
                    else:
                        specie = value[i:i + 2]
                        jump = 1
                else:
                    specie = value[i]
                    jump = 0
                j = 1
                number = ''
                while True:
                    if i + jump + j < len(value) and value[i + jump + j].isdigit():
                        number += value[i + jump + j]
                        j += 1
                    else:
                        break
                if number == '':
                    ret[specie] = 1
                else:
                    ret[specie] = int(number)
        return ret

    @staticmethod
    def formula_to_list(formula, nunits=1):
        """
        Reads a formula and returns a list of atomic symbols consistent with the formula and the number of
        formulas given by nunits

        :param formula: (str) Chemical formula as string
        :param nunits: (int) Number of formulas to apply
        :return: list of atomic symbols
        :rtype: list

        >>> Composition.formula_to_list('NaCl')
        ['Na', 'Cl']
        >>> flist = Composition.formula_to_list(u'Uut2Uup3Uus4Uuo5')
        >>> len(flist)
        14
        >>> flist = Composition.formula_to_list('Uut2Uup3Uus4Uuo5', nunits=2)
        >>> len(flist)
        28
        """
        # decompose composition
        a = re.findall(r"[A-Z][a-z0-9]*", formula)
        composition = []
        for i in a:
            m = re.match(r"([A-Za-z]+)([0-9]*)", i)
            if m.group(2) == "":
                n = int(1)
            else:
                n = int(m.group(2))

            for j in range(n * nunits):
                composition.append(m.group(1))

        return composition

    @property
    def gcd(self):
        """ Number of minimal formulas on a given composition.

        :return: The number of formulas that can be extracted from a composition ie, the greatest common denominator
                 for the composition.
        :rtype: int

        >>> cp = Composition('NaCl')
        >>> cp.gcd
        1
        >>> cp = Composition('Na2Cl2')
        >>> cp.gcd
        2
        >>> cp = Composition()
        >>> cp.gcd is None
        True
        """
        if self.natom > 0:
            return reduce(_gcd, self.values)
        else:
            return None

    @staticmethod
    def get_species_from_hex(arg):
        """List of species encoded for hex string produced by species_hex

        :return: Return a set of species from the encoded species created by the output of "species_hex" method.
        :param arg: str String with hexadecimal representation of list of species.

        >>> Composition.get_species_from_hex('0x38271d08')
        [8, 29, 39, 56]
        """
        num = int(arg, 16)
        ret = []
        while num > 0:
            ret.append(num % 256)
            num = (num-ret[-1])//256
        return ret

    @property
    def natom(self):
        """
        :return: The number of atoms in the composition
        :rtype: int

        >>> cp = Composition('H2O')
        >>> cp.natom
        3
        """
        return sum(self.values)

    @property
    def nspecies(self):
        """
        :return: Number of species in the composition
        :rtype: int

        >>> cp = Composition('H2O')
        >>> cp.nspecies
        2
        """
        return len(self.species)

    @property
    def symbols(self):
        """List of species on the composition

        :return: A list of atomic symbols
        :rtype: list

        >>> cp = Composition('H2O')
        >>> cp.symbols
        ['H', 'H', 'O']
        """
        ret = []
        for specie in self:
            number_atoms_specie = self.composition[specie]
            for i in range(number_atoms_specie):
                ret.append(specie)
        return sorted(deep_unicode(ret))

    @property
    def species(self):
        """List of species on the composition

        :return: The list of species, no particular order but atoms of the same specie are contiguous.
        :rtype: list

        >>> cp = Composition('H2O')
        >>> sorted(cp.species)
        ['H', 'O']
        """
        return [deep_unicode(x) for x in self._composition]

    def sorted_formula(self, sortby='alpha', reduced=True):
        """
        :return: The chemical formula. It could be sorted  alphabetically using sortby='alpha', by electronegativity
                 using sortby='electronegativity' or using Hill System with sortby='Hill'
                 Just the first 3 letters are unambiguous and case is not taken in account so you can use 'alp', 'hil'
                 or 'ele'
        :param sortby: (str) 'alpha' : Alphabetically
                             'electronegativity' : Electronegativity
                             'hill' : Hill System
        :param reduced: (bool) If the formula should be normalized
        :rtype: str

        .. notes: Hill exceptions have not being implemented yet

        >>> cp = Composition('YBa2Cu3O7')
        >>> cp.sorted_formula()
        'Ba2Cu3O7Y'
        >>> cp.sorted_formula(sortby='hill')
        'Ba2Cu3O7Y'
        >>> cp.sorted_formula(sortby='electroneg')
        'Ba2YCu3O7'
        >>> cp = Composition('H10C5')
        >>> cp.sorted_formula(sortby='hill', reduced=True)
        'CH2'
        >>> cp = Composition('IBr')
        >>> cp.sorted_formula(sortby='hill', reduced=False)
        'BrI'
        >>> cp = Composition('Cl4C')
        >>> cp.sorted_formula(sortby='hill', reduced=False)
        'CCl4'
        >>> cp = Composition('IH3C')
        >>> cp.sorted_formula(sortby='hill', reduced=False)
        'CH3I'
        >>> cp = Composition('BrH5C2')
        >>> cp.sorted_formula(sortby='hill', reduced=False)
        'C2H5Br'
        >>> cp = Composition('S04H2')
        >>> cp.sorted_formula(sortby='hill', reduced=False)
        'H2S4'
        >>> cp = Composition('SO4H2')
        >>> cp.sorted_formula(sortby='hill', reduced=False)
        'H2O4S'
        """
        if reduced and self.gcd > 1:
            comp = Composition(self.composition)
            for i in comp.composition:
                comp._composition[i] //= self.gcd
        else:
            comp = self
        if sortby.lower()[:3] == 'ele':
            electroneg = list(electronegativity(comp.species))
            # Not longer needed as electronegativy will return 0 for 'None' values
            # for i in range(len(electroneg)):
            #    if electroneg[i] is None:
            #        electroneg[i] = -1
            sortedspecies = array(comp.species)[argsort(electroneg)]
        elif sortby.lower()[:3] == "hil":  # FIXME: Hill system exceptions not implemented
            sortedspecies = []
            presortedspecies = sorted(comp.species)
            if 'C' in presortedspecies:
                sortedspecies.append('C')
                presortedspecies.pop(presortedspecies.index('C'))
            if 'H' in presortedspecies:
                sortedspecies.append('H')
                presortedspecies.pop(presortedspecies.index('H'))
            sortedspecies += presortedspecies
        else:
            sortedspecies = sorted(comp.species)
        ret = u''
        for specie in sortedspecies:
            ret += '%s' % specie
            if comp.composition[specie] > 1:
                ret += "%d" % comp.composition[specie]
        return deep_unicode(ret)

    def species_encoded(self, base):
        """Encode the list of species with a number

        :return: Encodes the species as a number.
        :param base: Integer used as base for encoding.
        :rtype: int

        >>> cp = Composition('H2O')
        >>> cp.species_encoded(100)
        801
        """
        ret = 0
        i = 0
        for atom_number in sorted(atomic_number(self.species)):
            ret += atom_number * (base ** i)
            i += 1
        return ret

    def species_hex(self):
        """Encoding in hexadecimal with 2 bytes per specie (base 256)

        :return: Encodes the species into a hexadecimal representation where each specie is stored on a 2-Byte slot
                 ordered by atomic number.
                 The output produces a unique encoding where each 2 character from the hexadecimal will encode a single
                 species and the species are ordered by atomic number making the codification unique.
        :rtype: str

        >>> cp = Composition('YBa2Cu3O7')
        >>> cp.species_hex()
        '0x38271d08'
        """
        enc = self.species_encoded(256)
        return hex(enc)

    @property
    def values(self):
        """
        :return: The number of atoms of each specie
        :rtype: list

        >>> cp = Composition('YBa2Cu3O7')
        >>> sorted(cp.values)
        [1, 2, 3, 7]
        """
        return [self._composition[x] for x in self._composition]





#!D:\bld\pychemia_1604228883768\_h_env\python.exe

import os
import sys
import json
import pychemia

filename = None
if len(sys.argv) == 2:
    filename = sys.argv[1]
else:
    print('Enter the JSON filename to store the data')
    exit(1)

dirs = [x for x in os.listdir('.') if os.path.isdir(x)]

ret = []
for idir in dirs:
    if os.path.isfile(idir + '/POSCAR'):
        try:
            st = pychemia.code.vasp.read_poscar(idir + '/POSCAR')
        except ValueError:
            print('Bad POSCAR\n\n' + open(idir + '/POSCAR').read())
            continue
        # shutil.copy2(idir+'/POSCAR',idir+'_POSCAR')
        print(idir, st.natom)
    else:
        st = pychemia.Structure.load_json(idir + '/structure.json')
        # shutil.copy2(idir+'/structure.json',idir+'_structure.json')
        print('ERROR:', idir, st.natom)
        continue

    if os.path.isfile(idir + '/OUTCAR'):
        try:
            vo = pychemia.code.vasp.VaspOutput(idir + '/OUTCAR')
        except ValueError:
            print('Error reading Vasp Output @ ' + idir + '/OUTCAR')
            continue
        if not vo.has_forces_stress_energy():
            print('Error extracting forces @ ' + idir + '/OUTCAR')
            continue
    else:
        print('No OUTCAR found @ ' + idir)
        continue

    spacegroup = pychemia.crystal.CrystalSymmetry(st).number()
    energy_pa = vo.final_data['energy']['free_energy'] / st.natom

    data = {'id': idir, 'energy_pa': energy_pa, 'natom': st.natom, 'spacegroup': spacegroup,
            'forces': vo.relaxation_info()['avg_force'],
            'stress': max(vo.relaxation_info()['avg_stress_diag'], vo.relaxation_info()['avg_stress_non_diag'])}

    ret.append(data)

wf = open(filename, 'w')
json.dump(ret, wf, sort_keys=True, indent=4, separators=(',', ': '))
wf.close()




from numbers import Number
from math import sqrt
import gzip
import hashlib
import os
import sys
import zipfile


def deep_unicode(value):
    """
    Recursively convert to unicode all string-like elements of a complex object.
    Use with care for objects that could not be converted properly to string
    This is safe for things like Atom names and symbols

    :param value: (unicode, list, dict)
    :return: (str, list, dict) The string, list or dictionary where all string-like elements are converted into strings

    :rtype : (str, list, dict)

    Examples:
    >>> deep_unicode(u'abc')
    u'abc'
    >>> deep_unicode([u'abc'])
    [u'abc']
    >>> deep_unicode({u'abc': u'def'})
    {u'abc': u'def'}
    >>> deep_unicode('abc')
    u'abc'

    """
    if hasattr(value, 'decode'):
        value = value.decode()
    if isinstance(value, dict):
        ret = {}
        for key in value:
            ret[deep_unicode(key)] = deep_unicode(value[key])
        return ret
        # This line is Python 2.7+
        # return {unicode2string(key): unicode2string(value) for key, value in value.items()}
    elif isinstance(value, str):
        return value
    elif isinstance(value, Number):
        return value
    elif value is None:
        return None
    else:
        return value


def convert_color(s):
    """
    Convert a string hexadecimal representation of a color into a tuple
    RBG where each element of the tuple is between 0.0 to 1.0

    :param s: (str)
    :return: (tuple) With 3 floats representing the color in RGB

    :rtype : tuple

    Examples:
    >>> convert_color('FF5500')
    (1.0, 0.3333333333333333, 0.0)

    """
    return float(int(s[:2], 16)) / 255, float(int(s[2:4], 16)) / 255, float(int(s[4:6], 16)) / 255


def get_int(value):
    """
    Convert a string to an integer

    :param value: (str)
    :return: (int)

    Example:
    >>> get_int('3')
    3

    """
    if value.isdigit():
        return int(value)
    else:
        print("ERROR: The value '%s' should be an integer" % value)
        sys.exit(2)


def get_float(value):
    """
    Convert a string to a float number

    :param value: (str)
    :return: (float)

    Example:
    >>> get_float('3.0')
    3.0

    """
    try:
        ret = float(value)
    except ValueError:
        raise ValueError("Could not convert '%s' into a float number" % value)
    return ret


def hashfile(filename):
    """
    Get the MD5 hash sum of a given file

    :param filename: (str)
    :return: (str)

    >>> import tempfile
    >>> a = tempfile.NamedTemporaryFile()
    >>> hashfile(a.name)
    'd41d8cd98f00b204e9800998ecf8427e'
    >>> a = tempfile.NamedTemporaryFile('w')
    >>> tmp= a.file.write(128000*'GAF')
    >>> a.file.flush()
    >>> hashfile(a.name)
    '7b8a4f8a3ce222580765d577df78b782'

    """
    blocksize = 65536
    hasher = hashlib.md5()
    with open(filename, 'rb') as afile:
        buf = afile.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(blocksize)
    return hasher.hexdigest()


def read_file(filename):
    """
    General function to open files even if they are
    compressed

    :param filename:
    :return:

    Example:
    >>> import tempfile
    >>> a = tempfile.NamedTemporaryFile()
    >>> read_file(a.name)
    ''

    """
    if not os.path.exists(filename):
        raise ValueError('Could not open file: %s' % filename)

    if filename[-3:] == '.gz':
        rf = gzip.GzipFile(filename)
    elif filename[-4:] == '.zip':
        rf = zipfile.ZipFile(filename)
    else:
        rf = open(filename)
    return rf.read()


def only_ascii(string):
    """
    Remove non-ascii characters, from monty

    :param string:
    :return:

    Example:
    >>> only_ascii(u'lmnopq')
    u'lmnopq'

    """
    return "".join(x for x in string if ord(x) < 128)


def string2number(word):
    """
    Take a string and try to convert into a number
    of list of numbers.
    It try to returns integer values always that is
    is possible.

    Examples:
              '3'         =>     3
              '3.14'      =>     3.14
              '1.5e-5'    =>     0.000015
              '1.5d-5'    =>     0.000015
              '1.5E-5'    =>     0.000015
              '1.5D-5'    =>     0.000015
              '1/2'       =>     0.5
              'sqrt(3.0)' =>     1.732050
    Args:
        word:
           An string that should be converted into a number

    Returns:
        number:
           The value extracted
        kind:
           The kind of value

    """
    number = None
    kind = None
    try:
        number = int(word)
        kind = 'int'
    except ValueError:
        try:
            number = float(word)
            kind = 'float'
        except ValueError:
            if 'd' in word:
                word = word.replace('d', 'e')
            elif 'D' in word:
                word = word.replace('D', 'e')
            try:
                number = float(word)
                kind = 'float'
            except ValueError:
                if '/' in word:
                    splt = word.split('/')

                    if splt[0].isdigit() and splt[1].isdigit():
                        number = float(splt[0]) / float(splt[1])
                        kind = 'float'
                elif word[:4] == 'sqrt':
                    number = sqrt(float(word[5:-1]))
                    kind = 'float'

                elif word[:5] == '-sqrt':
                    number = -sqrt(float(word[6:-1]))
                    kind = 'float'
                else:
                    print('ERROR: "%s" is not a number\n' % word)
    return number, kind



from numpy import sum
from .searcher import Searcher
from pychemia import pcm_log

"""
Implementation of Genetic Algorithms
This is the abstract layer
No references to Structural Search should be appear
This module should be general enough even for stock market prediction!
"""

__author__ = 'Guillermo Avendano-Franco'


class GeneticAlgorithm(Searcher):
    def __init__(self, population, params=None, generation_size=32, stabilization_limit=10):
        Searcher.__init__(self, population, generation_size, stabilization_limit)
        self.nelite = None
        self.mutation_prob = None
        self.cross_prob = None
        self.crossing_sets = None
        self.set_params(params)

    def set_params(self, params):
        self.nelite = 2
        self.mutation_prob = 0.5
        self.cross_prob = 0.5
        if self.generation_size == 32:
            self.crossing_sets = [10, 5]
        elif self.generation_size == 16:
            self.crossing_sets = [5, 2]
        elif self.generation_size == 8:
            self.crossing_sets = [2, 1]

        if params is None:
            params = {}
        if 'nelite' in params:
            self.nelite = params['nelite']
        if 'mutation_prob' in params:
            self.mutation_prob = params['mutation_prob']
        if 'cross_prob' in params:
            self.cross_prob = params['cross_prob']
        if 'crossing_sets' in params:
            self.crossing_sets = params['crossing_sets']

        self.nelite = int(self.generation_size - 2 * sum(self.crossing_sets))
        if self.nelite <= 0:
            raise ValueError('Enter explicit crossing_sets for your generation size')

    def get_params(self):
        return {'nelite': self.nelite, 'mutation_prob': self.mutation_prob, 'cross_prob': self.cross_prob,
                'crossing_sets': self.crossing_sets}

    def run_one(self):
        # Get a static selection of the values in the generation that are relaxed
        selection = self.population.ids_sorted(self.actives_in_generation)
        pcm_log.info(' Size of selection : %d' % len(selection))

        # Minus sign because we are searching for minima
        intensity = self.population.get_values(selection)
        for entry_id in intensity:
            intensity[entry_id] *= -1

        discarded_index = self.nelite + int(sum(self.crossing_sets))

        for i in range(min(self.nelite, len(selection))):
            entry_id = selection[i]
            pcm_log.debug('[%s] Promoted to new generation' % str(entry_id))
            self.pass_to_new_generation(entry_id, reason='Elite')

        jump = 0
        for i in range(min(len(self.crossing_sets), len(selection))):
            for j in range(self.crossing_sets[i]):
                entry_id = selection[i]

                if self.nelite + j + jump < len(selection):
                    entry_jd = selection[self.nelite + j + jump]
                    new_entry_id, new_entry_jd = self.population.cross([entry_id, entry_jd])
                    msg = 'Replace candidates %d and %d by crossing %d with %d'
                    pcm_log.info(msg % (self.nelite + j + jump, discarded_index, self.nelite + j + jump, i))

                    pcm_log.debug('[%s] Moved to: %s' % (entry_id, new_entry_id))
                    self.replace_by_other(entry_jd, new_entry_id,
                                          reason='Cross between %s and %s' % (entry_id, entry_jd))
                    self.population.enable(new_entry_id)
                    if discarded_index < len(selection):
                        entry_kd = selection[discarded_index]
                        pcm_log.debug('[%s] Moved to: %s' % (entry_kd, new_entry_jd))
                        self.replace_by_other(entry_kd, new_entry_jd,
                                              reason='Cross between %s and %s' % (entry_id, entry_jd))
                        self.population.enable(new_entry_jd)
                    else:
                        pcm_log.debug('Candidate %s will not be activated' % new_entry_jd)

                discarded_index += 1
            jump += self.crossing_sets[i]



from ..codes import CodeRun
from .input import AbinitInput
import os


class AbinitRun(CodeRun):

    def __init__(self, workdir):
        CodeRun.__init__(self, executable='abinit', workdir=workdir, use_mpi=True)
        self.stdout_filename = 'abinit.out'
        self.stderr_filename = 'abinit.err'
        self.stdin_filename = 'abinit.files'

    def set_inputs(self):
        if not os.path.isfile(self.workdir+os.sep+'abinit.in'):
            raise ValueError("Input file 'abinit.in' not found at: %s" % self.workdir)
        else:
            self.input_path = self.workdir + os.sep + 'abinit.in'
            self.input = AbinitInput(self.input_path)

    def get_outputs(self):
        pass

    def is_finished(self):
        pass
    
    def is_successful(self):
        pass



from html.parser import HTMLParser
import sys


class MyHTMLParser(HTMLParser):
    """
    Create a subclass and override the handler methods
    """

    def error(self, message):
        pass

    def handle_starttag(self, tag, attrs):
        sys.stdout.write("'")

    def handle_endtag(self, tag):
        sys.stdout.write("'")

    def handle_data(self, data):
        sys.stdout.write(data)



from ..codes import CodeRun
from .input import OctopusInput
import os


class OctopusRun(CodeRun):

    def __init__(self, workdir='.'):
        CodeRun.__init__(self, executable='octopus', workdir=workdir, use_mpi=True)
        self.stdout_filename = 'octopus.out'
        self.stderr_filename = 'octopus.err'

    def set_inputs(self):
        if not os.path.isfile(self.workdir+os.sep+'inp'):
            raise ValueError("Input file 'inp' not found at: %s" % self.workdir)
        else:
            self.input_path = self.workdir + os.sep + 'inp'
            self.input = OctopusInput(self.input_path)

    def get_outputs(self):
        pass

    def is_finished(self):
        pass
    
    def is_successful(self):
        pass




# THIS FILE IS GENERATED FROM PYCHEMIA SETUP.PY
short_version = '0.20.6.1'
version = '0.20.6.1'
full_version = '0.20.6.1.dev0+Unknown'
git_revision = 'Unknown'
name = 'pychemia'
description = 'Python Framework for Materials Discovery and Design'
url = 'https://materialsdiscovery.github.io/PyChemia'
author = 'Guillermo Avendaño Franco'
email = 'gufranco@mail.wvu.edu'
status = 'Development'
copyright = 'Copyright 2020'
date = 'June 1, 2020'
release = False
if not release:
    version = full_version



import os
import numpy as np


def plot_energy(path):
    import matplotlib.pylab as plt
    if not os.path.isdir(path):
        print('ERROR: No such dir ', path)
        return None
    if not os.path.isfile(path + '/td.general/energy'):
        print('ERROR: No energy file ', path + '/td.general/energy')
        return None

    data = np.loadtxt(path + '/td.general/energy')

    if len(data) == 0:
        print('No data on ', path + '/td.general/energy')
        return None

    plt.clf()
    fig = plt.figure(num=None, figsize=(16, 12), dpi=80, facecolor='w', edgecolor='k')
    labels = ['Total', 'Kinetic (ions)', 'Ion-Ion', 'Electronic',
              'Eigenvalues', 'Hartree', 'Int[n v_xc]', 'Exchange', 'Correlation']

    # For some columns the energy is only computed each
    # certain steps, this will identify the number of
    # steps to get a new non-zero value
    nstep = 0
    for i in range(len(data)):
        if i == 0 and data[i, -3] != 0.0:
            nstep += 1
        if data[i, -3] == 0.0:
            nstep += 1
        if i != 0 and data[i, -3] != 0.0:
            break
    print('Complete data each', nstep, ' steps')

    for i in range(9):
        plt.subplot(330 + i + 1)
        plt.plot(data[::nstep, 1], data[::nstep, i + 2], label=labels[i])
        plt.xlabel('Time [hbar/H]')
        plt.legend(loc=4)

    return fig



import math
import sys

import numpy as np

if 'matplotlib' not in sys.modules:
    import matplotlib

    matplotlib.use('agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from pychemia.code.abinit import AbinitInput, AbiFiles
from pychemia.utils.periodic import covalent_radius
from pychemia.utils.constants import bohr_angstrom


def compute_bonds(typat, xcart, znucl):
    """
    Compute the bond lengths of all the atoms
    inside the unitary box (NEEDS EXTENSION TO
    OUTSIDE THE BOX)

    :param typat: (int, list) Type of atoms
    :param xcart: (numpy.ndarray) Cartesian positions
    :param znucl: (int, list) Atomic number for atoms in typat
    :return:
    """
    print(typat)
    print(xcart)
    print(znucl)

    npxcart = np.array(xcart).reshape((-1, 3))
    if isinstance(typat, int):
        lsttypat = [typat]
    else:
        lsttypat = typat
    if isinstance(znucl, (int, float)):
        lstznucl = [znucl]
    else:
        lstznucl = znucl
    covrad = [covalent_radius(lstznucl[iatom - 1]) for iatom in lsttypat]

    bonds = []
    for iatom in range(len(npxcart)):
        for jatom in range(iatom + 1, len(npxcart)):
            # Compute bond length between atoms i and j
            bl = math.sqrt(sum((npxcart[jatom] - npxcart[iatom]) ** 2))
            if 1.35 * (covrad[iatom] + covrad[jatom]) > bl:
                bonds.append([iatom, jatom, bl, (npxcart[jatom] - npxcart[iatom]) / bl])
            else:
                print("small distance: %s" % bl)

    return bonds


def save_bonds(allbonds, abivar):
    wf = open("bonds.txt", 'w')

    for ii in range(len(allbonds[0])):  # Number of bonds
        label = abivar.atom_name(allbonds[0][ii][0]) + ' ' + abivar.atom_name(allbonds[0][ii][1])
        wf.write(label)
        for jj in range(len(allbonds)):  # Number of sets
            wf.write(' ' + str(round(bohr_angstrom * allbonds[jj][ii][2], 4)).ljust(10))  # Bond length
        wf.write('\n')
    wf.close()


def get_all_bonds(abinitfile, dtset=''):
    allbonds = []
    index = 0
    for abf in abinitfile:
        if dtset == '':
            idt = ''
        else:
            idt = int(dtset[index])
            index += 1
        filep = abf.basedir + "/" + abf.files['tmpout'] + "_OUT.nc"

        out = AbinitInput(filep)
        
        print(out.variables)

        xcart = out.get_value('xcart', idtset=idt)
        znucl = out.get_value('znucl', idtset=idt)
        typat = out.get_value('typat', idtset=idt)

        bonds = compute_bonds(typat, xcart, znucl)
        allbonds.append(bonds)

    abivar = AbinitInput(abinitfile[0].get_input_filename())
    save_bonds(allbonds, abivar)
    return allbonds


def plot_bonds(listabifile, listidtset):
#    print(listabifile)
#    print(listidtset)
    pp = PdfPages("bonds.pdf")
    plt.ioff()

    if not listidtset:
        allbonds = get_all_bonds(listabifile)
    else:
        allbonds = get_all_bonds(listabifile, listidtset)
    print(allbonds)

    plt.figure(figsize=(32, 20), dpi=100)
    plt.subplots_adjust(left=0.05, bottom=0.05, right=0.95, top=0.95,
                        wspace=None, hspace=None)

    abivar = AbinitInput(listabifile[0].get_input_filename())

    nbond = len(allbonds[0])
    n = 0
    for ibond in allbonds:
        if len(ibond) != nbond:
            print('ERROR: Number of bonds changed')
            if not listidtset:
                print(listabifile[n].basedir, 'contains', len(ibond), 'bonds')
            else:
                print(listabifile[n].filename + ':' + listidtset[n], 'contains', len(ibond), 'bonds')
                n += 1

    iplot = 1
    for ibond in range(nbond):
        plt.subplot(5, 4, iplot)
        y = [bohr_angstrom * seti[ibond][2] for seti in allbonds]
        print(abivar.atom_name(seti[ibond][0]))
        print(abivar.atom_name(seti[ibond][1]))
        label = abivar.atom_name(seti[ibond][0]) + ' ' + abivar.atom_name(seti[ibond][1])
        plt.plot(y, label=label)
        plt.plot(y, 'ro')
        iplot += 1
        plt.legend()

    pp.savefig()
    plt.clf()

    for ibond in range(nbond):
        y = [bohr_angstrom * seti[ibond][2] for seti in allbonds]
        label = abivar.atom_name(seti[ibond][0]) + ' ' + abivar.atom_name(seti[ibond][1])
        plt.plot(y, label=label)
        plt.plot(y, 'ro')

        plt.text(0.09, y[0] + 0.001, label, size='small')
        iplot += 1
        plt.legend()

    pp.savefig()
    pp.close()


if __name__ == '__main__':

    narg = len(sys.argv) - 1

    list_abifile = []
    list_idtset = []
    dtsetall = False
    for i in range(narg):
        filename = sys.argv[i + 1]
        if ':' in filename:
            idtset = filename.split(':')[1]
            if idtset == 'all':
                dtsetall = True
            else:
                list_idtset.append(idtset)
            filename = filename.split(':')[0]

        abifile = AbiFiles(filename)
        if dtsetall:
            av = AbinitInput(abifile.get_input_filename())
            keys = av.get_dtsets_keys()
#            print(keys)
            for j in keys:
                list_abifile.append(abifile)
                list_idtset.append(str(j))

        else:
            list_abifile.append(abifile)

    plot_bonds(list_abifile, list_idtset)



#!D:\bld\pychemia_1604228883768\_h_env\python.exe

import os
import json
import sys
import time
import argparse
import multiprocessing
import pychemia
from pychemia.db import get_database


def safe_read_json(filename):
    """
    Safely read a given filename, extract and returns the associated dictionary
    """
    if not os.path.exists(filename):
        raise ValueError("ERROR: Could not read file: %s" % filename)
    rf = open(filename)
    try:
        data = json.load(rf)
    except ValueError:
        raise ValueError("ERROR: File is not in proper JSON format: %s" % filename)
    rf.close()
    return data


def searcher(dbsettings, popu_settings, searcher_settings, composition, max_formulas):

    pcdb = get_database(dbsettings)

    if 'target_forces' in popu_settings:
        target_forces = popu_settings['target_forces']
    else:
        target_forces = 1E-3
    if 'value_tol' in popu_settings:
        value_tol = popu_settings['value_tol']
    else:
        value_tol = 1E-2
    if 'distance_tolerance' in popu_settings:
        distance_tolerance = popu_settings['distance_tolerance']
    else:
        distance_tolerance = 0.3

    # Population
    popu = pychemia.population.RelaxStructures(pcdb, composition=composition, tag=composition,
                                               target_forces=target_forces, value_tol=value_tol,
                                               distance_tolerance=distance_tolerance, min_comp_mult=1,
                                               max_comp_mult=max_formulas)

    # Global Searcher
    if 'target_value' in searcher_settings:
        target_value = searcher_settings['target_value']
    else:
        target_value = None

    method = searcher_settings['method']
    if method == 'FireFly':
        globsearcher = pychemia.searcher.FireFly(popu, params=searcher_settings['params'],
                                                 generation_size=searcher_settings['generation_size'],
                                                 stabilization_limit=searcher_settings['stabilization_limit'],
                                                 target_value=target_value)

    globsearcher.run()


def manager(population, code):
    pass


def worker(population, code, candidate):
    pass


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='PyChemia Structural Searcher')

    subparsers = parser.add_subparsers(help='commands', dest='subparser_name')

    # The Searcher uses global searchers, one for each composition 
    searcher_parser = subparsers.add_parser('searcher', help='Create one searcher for each composition')

    searcher_parser.add_argument('-clean', action='store_true',
                                 help='If true, clean the entire database before populate',
                                 required=False, default=False)
    searcher_parser.add_argument('--db_settings', '-d', type=str,
                                 help='Path to JSON file with database settings',
                                 required=True)
    searcher_parser.add_argument('--population_settings', '-p', type=str,
                                 help='Path to JSON file with population settings',
                                 required=True)
    searcher_parser.add_argument('--searcher_settings', '-s', type=str,
                                 help='Path to JSON file with searcher settings',
                                 required=True)
    searcher_parser.add_argument('-compositions', '-c', type=str,
                                 help='Path to JSON file with settings for all compositions to search',
                                 required=True)

    # The Manager search the database for unevaluated candidates, create submission jobs for them and 
    # collect results after execution
    manager_parser = subparsers.add_parser('manager', help='Search a database for candidates to be evaluated')

    manager_parser.add_argument('-db_settings', '-d', type=str,
                                help='Path to JSON file with database settings',
                                required=True)
    manager_parser.add_argument('-population_settings', '-p', type=str,
                                help='Path to JSON file with settings relative to relaxation',
                                required=True)
    manager_parser.add_argument('-basepath', type=str,
                                help='Path where calculations are performed',
                                required=False, default='.')

    # The Worker takes one candidate and execute the selected code to relax the structure
    worker_parser = subparsers.add_parser('worker', help='Takes one candidate and compute a local relaxation on it')

    worker_parser.add_argument('-case', type=str,
                               help='ID of candidate to relax',
                               required=True)
    worker_parser.add_argument('-basepath', type=str,
                               help='Path where calculations are performed',
                               required=False, default='.')

    args = parser.parse_args()
    print(args)

    if args.subparser_name in ['searcher', 'manager']:

        db = safe_read_json(args.db_settings)
        pcdb = get_database(db) 

        print("")
        print("Database Settings")
        print("-----------------\n")
        print(pcdb)
        print("")

        popu_settings = safe_read_json(args.population_settings)
        print("Population Settings")
        print("-------------------\n")
        for i in popu_settings:
            print(" %30s : %s" % (i.ljust(30), popu_settings[i]))
        print("")

    if args.subparser_name == 'searcher':

        searcher_settings = safe_read_json(args.searcher_settings)
        print("Searcher Settings")
        print("-----------------\n")
        for i in searcher_settings:
            print(" %30s : %s" % (i.ljust(30), searcher_settings[i]))
        print("")

        comp_settings = safe_read_json(args.compositions)
        print("Compositions and max number of formulas allowed")
        print("---------------------------------------------\n")
        for i in comp_settings:
            print(" %30s : %s" % (i.ljust(30), comp_settings[i]))
        print("")

        # Check consistency of searcher settings
        if 'method' not in searcher_settings:
            print('Not "method" was declared in searcher settings')
            sys.exit(1)
        if 'generation_size' not in searcher_settings:
            print('Not "generation_size" was declared in searcher settings')
            sys.exit(1)
        if 'stabilization_limit' not in searcher_settings:
            print('Not "stabilization_limit" was declared in searcher settings')
            sys.exit(1)

        sprocs = {}
        for composition in comp_settings:
            max_formulas = comp_settings[composition]
            sprocs[composition] = multiprocessing.Process(target=searcher,
                                                          args=(db,
                                                                popu_settings,
                                                                searcher_settings,
                                                                composition,
                                                                max_formulas))
            sprocs[composition].start()
        
    elif args.subparser_name == 'manager':

        popu = pychemia.population.RelaxStructures(pcdb)
        print(popu)

        while True:
            
            for icase in popu.members:
                if not popu.is_evaluated(icase):
                    ipath = args.basepath + os.sep + str(icase)
                    if not os.path.isdir(ipath):
                        os.mkdir(ipath)

                    st = popu.get_structure(icase) 
                    task = pychemia.code.abinit.task.IonRelaxation(st)

